#!/usr/bin/env python
"""
findSpots.py -- detect spots in 2D images

"""
# Numeric
import numpy as np

# Real-valued FFT
from numpy.fft import rfft2, irfft2, fftshift

# Image processing / filtering
from scipy import ndimage as ndi

# Caching
from functools import lru_cache

# Custom utilities
from .helper import (
    pad,
    threshold_image,
    stable_divide_array,
    assign_methods,
    hollow_box_var,
)


def gauss(I, k=1.0, w=9, t=200.0, return_filt=False):
    """
    Convolve the image with a simple Gaussian kernel, then apply
    a static threshold

    args
    ----
        I           :   2D ndarray (YX)
        k           :   float, kernel sigma
        w           :   int, kernel window size
        t           :   float, threshold
        return_filt :   bool, also return the filtered image
                        and boolean image

    returns
    -------
        if return_filt:
        (
            2D ndarray, the post-convolution image;
            2D ndarray, the thresholded binary image;
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot
        )
        else
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot

    """
    # Compute the transfer function
    G_rft = _gauss_setup(*I.shape, k, w)
    return threshold_image(
        fftshift(irfft2(rfft2(I) * G_rft, s=I.shape)),
        t=t,
        return_filt=return_filt,
    )


@lru_cache(maxsize=1)
def _gauss_setup(H, W, k, w):
    """
    Make the transfer function for Gaussian filtering
    of real-valued images.

    args
    ----
        H, W    :   int, height and width of image
        k       :   float, kernel sigma
        w       :   int, window size for the kernel

    returns
    -------
        2D ndarray, dtype complex128, the RFFT2 of
            the Gaussian kernel

    """
    S = 2 * (k**2)
    g = np.exp(-((np.indices((w, w)) - (w - 1) / 2) ** 2).sum(0) / S)
    return rfft2(pad(g / g.sum(), H, W))


def centered_gauss(I, k=1.0, w=9, t=200.0, return_filt=False):
    """
    Convolve the image with a mean-subtracted Gaussian kernel,
    then apply a static threshold.

    args
    ----
        I           :   2D ndarray (YX)
        k           :   float, kernel sigma
        w           :   int, kernel window size
        t           :   float, threshold
        return_filt :   bool, also return the filtered image
                        and boolean image

    returns
    -------
        if return_filt:
        (
            2D ndarray, the post-convolution image;
            2D ndarray, the thresholded binary image;
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot
        )
        else
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot

    """
    # Compute the transfer function
    G_rft = _centered_gauss_setup(*I.shape, k, w)
    return threshold_image(
        fftshift(irfft2(rfft2(I) * G_rft, s=I.shape)),
        t=t,
        return_filt=return_filt,
    )


@lru_cache(maxsize=1)
def _centered_gauss_setup(H, W, k, w):
    """
    Make the transfer function for convolution of real-valued
    images with a centered (mean-subtracted) Gaussian kernel.

    args
    ----
        H, W    :   int, height and width of image
        k       :   float, kernel sigma
        w       :   int, window size for the kernel

    returns
    -------
        2D ndarray, dtype complex128, the RFFT2 of
            the kernel

    """
    S = 2 * (k**2)
    g = np.exp(-((np.indices((w, w)) - (w - 1) / 2) ** 2).sum(0) / S)
    g = g / g.sum()
    return rfft2(pad(g - g.mean(), H, W))


def mle_amp(I, k=1.0, w=9, t=200.0, return_filt=False):
    """
    Convolve the image with a mean-subtracted Gaussian kernel,
    square the result and normalize by the then apply a static threshold. This is equivalent to

    args
    ----
        I           :   2D ndarray (YX)
        k           :   float, kernel sigma
        w           :   int, kernel window size
        t           :   float, threshold
        return_filt :   bool, also return the filtered image
                        and boolean image

    returns
    -------
        if return_filt:
        (
            2D ndarray, the post-convolution image;
            2D ndarray, the thresholded binary image;
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot
        )
        else
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot

    """
    # Compute the transfer function and normalization factor
    G_rft, Sgc2 = _mle_amp_setup(*I.shape, k, w)

    # Perform filtering
    return threshold_image(
        (fftshift(irfft2(rfft2(I) * G_rft, s=I.shape)) ** 2) / Sgc2,
        t=t,
        return_filt=return_filt,
    )


@lru_cache(maxsize=1)
def _mle_amp_setup(H, W, k, w):
    """
    Make the transfer function for convolution of real-valued
    images with a centered (mean-subtracted) Gaussian kernel,
    along with a normalization term to convert to the max-likelihood
    estimate for the intensities of Gaussian spots.

    args
    ----
        H, W    :   int, height and width of image
        k       :   float, kernel sigma
        w       :   int, window size for the kernel

    returns
    -------
        (
            2D ndarray, dtype complex128, the RFFT2 of
                the kernel
            float, normalization factor
        )

    """
    S = 2 * (k**2)
    g = np.exp(-((np.indices((w, w)) - (w - 1) / 2) ** 2).sum(0) / S)
    g = g / g.sum()
    gc = g - g.mean()
    Sgc2 = (gc**2).sum()
    return rfft2(pad(gc, H, W)), Sgc2


def dog(I, k0=1.0, k1=3.0, w=9, t=200.0, return_filt=False):
    """
    Convolve the image with a difference-of-Gaussians kernel,
    then apply a static threshold.

    args
    ----
        I           :   2D ndarray
        k0          :   float, positive kernel sigma
        k1          :   float, negative kernel sigma
        w           :   int, kernel size
        t           :   float, threshold
        return_filt :   bool, also return the filtered image

    returns
    -------
        if return_filt:
        (
            2D ndarray, the post-convolution image;
            2D ndarray, the thresholded binary image;
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot
        )
        else
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot

    """
    # Generate the transfer function
    dog_tf = _dog_setup(*I.shape, k0, k1, w)

    # Perform the convolution
    return threshold_image(
        fftshift(irfft2(rfft2(I) * dog_tf, s=I.shape)),
        t=t,
        return_filt=return_filt,
    )


@lru_cache(maxsize=1)
def _dog_setup(H, W, k0, k1, w):
    """
    Generate the transfer function for DoG filtering.

    args
    ----
        H, W        :   int, height and width of the image
                        to be convolved
        k0          :   float, positive kernel sigma
        k1          :   float, negative kernel sigma
        w           :   int, kernel size

    returns
    -------
        2D ndarray, RFFT of the DoG kernel

    """
    S0 = 2 * (k0**2)
    S1 = 2 * (k1**2)
    g1 = np.exp(
        -((np.indices((int(w), int(w))) - (int(w) - 1) / 2) ** 2).sum(0) / S0
    )
    g1 = g1 / g1.sum()
    g2 = np.exp(
        -((np.indices((int(w), int(w))) - (int(w) - 1) / 2) ** 2).sum(0) / S1
    )
    g2 = g2 / g2.sum()
    return rfft2(pad(g1 - g2, H, W))


def log(I, k=1.0, w=11, t=200.0, return_filt=False):
    """
    Detect spots by Laplacian-of-Gaussian filtering.

    args
    ----
        I           :   2D ndarray
        k           :   float, kernel sigma
        w           :   int, kernel size
        t           :   float, threshold
        return_filt :   bool, also return the filtered image

    returns
    -------
        if return_filt:
        (
            2D ndarray, the post-convolution image;
            2D ndarray, the thresholded binary image;
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot
        )
        else
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot

    """
    # Generate the transfer function
    G_rft = _log_setup(*I.shape, k, w)

    # Perform the convolution
    return threshold_image(
        fftshift(irfft2(rfft2(I) * G_rft, s=I.shape)),
        t=t,
        return_filt=return_filt,
    )


def _log_setup(H, W, k, w):
    """
    Generate a Laplacian-of-Gaussian (LoG) transfer function
    for subsequent convolution with an image.

    args
    ----
        H, W    :   int, height and width of target image
        k       :   float, kernel sigma
        w       :   int, kernel size

    returns
    -------
        2D ndarray, dtype complex128, the transfer function

    """
    S = 2 * (k**2)
    g = np.exp(-((np.indices((w, w)) - (w - 1) / 2) ** 2).sum(0) / S)
    g = g / g.sum()
    log_k = -ndi.laplace(g)
    return rfft2(pad(log_k, H, W))


def dou(I, w0=3, w1=11, t=200.0, return_filt=False):
    """
    Uniform-filter an image with two different kernel sizes and
    subtract them. This is like a poor-man's version of DoG
    filtering.

    args
    ----
        I           :   2D ndarray
        w0          :   int, positive kernel size
        w1          :   int, negative kernel size
        t           :   float, threshold
        return_filt :   bool, also return the filtered image

    returns
    -------
        if return_filt:
        (
            2D ndarray, the post-convolution image;
            2D ndarray, the thresholded binary image;
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot
        )
        else
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot

    """
    return threshold_image(
        ndi.uniform_filter(I, w0) - ndi.uniform_filter(I, w1),
        t=t,
        return_filt=return_filt,
    )


def min_max(I, w=9, t=200.0, mode="constant", return_filt=False, **kwargs):
    """
    Use the difference between the local maximum and local minimum
    in square subwindows to identify spots in an image.

    args
    ----
        I           :   2D ndarray
        w           :   int, window size for test
        t           :   float, threshold for spot detection
        mode        :   str, behavior at boundaries (see scipy.ndimage
                        documentation)
        kwargs      :   to ndimage.maximum_filter/minimum_filter

    returns
    -------
        if return_filt:
        (
            2D ndarray, the post-convolution image;
            2D ndarray, the thresholded binary image;
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot
        )
        else
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot

    """
    size = (w, w)
    I_filt = I - ndi.minimum_filter(I, size=size, mode=mode, **kwargs)

    # Set the probability of detection near the border to zero
    hw = w // 2
    I_filt[:hw, :] = t - 1
    I_filt[:, :hw] = t - 1
    I_filt[-hw:, :] = t - 1
    I_filt[:, -hw:] = t - 1

    # Threshold the result
    return threshold_image(I_filt, t=t, return_filt=return_filt)


def gauss_filt_min_max(
    I, k=1.0, w=9, t=200.0, mode="constant", return_filt=False, **kwargs
):
    """
    Similar to min_max_filter(), but perform a Gaussian convolution
    prior to looking for spots by min/max filtering.

    args
    ----
        I           :   2D ndarray
        k           :   float, Gaussian kernel sigma
        w           :   int, window size for test
        t           :   float, threshold for spot detection
        mode        :   str, behavior at boundaries (see scipy.ndimage
                        documentation)
        kwargs      :   to ndimage.maximum_filter/minimum_filter

    returns
    -------
        if return_filt:
        (
            2D ndarray, the post-convolution image;
            2D ndarray, the thresholded binary image;
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot
        )
        else
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot

    """
    # Generate the kernel for convolution
    G_rft = _gauss_setup(*I.shape, k, w)

    # Perform the convolution and do min/max filtering on
    # the result
    return min_max(
        fftshift(irfft2(rfft2(I) * G_rft, s=I.shape)),
        w=w,
        t=t,
        mode=mode,
        return_filt=return_filt,
        **kwargs
    )


def llr(I, k=1.0, w=9, t=20.0, return_filt=False):
    """
    Perform a log-likelihood ratio test for the presence of
    spots. This is the ratio of likelihood of a Gaussian spot
    in the center of the subwindow, relative to the likelihood
    of flat background with Gaussian noise.

    args
    ----
        I           :   2D ndarray
        k           :   float, Gaussian kernel sigma
        w           :   int, window size for test
        t           :   float, threshold for spot detection
        return_filt :   bool, also return filtered image

    returns
    -------
        if return_filt:
        (
            2D ndarray, the post-convolution image;
            2D ndarray, the thresholded binary image;
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot
        )
        else
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot

    """
    # Generate the convolution kernel and normalization factor
    G_rft, Sgc2 = _mle_amp_setup(*I.shape, k, w)
    n_pixels = w**2

    # Perform the convolutions for detection
    A = ndi.uniform_filter(I, w)
    B = ndi.uniform_filter(I**2, w)
    C = fftshift(irfft2(rfft2(I) * G_rft, s=I.shape)) ** 2

    # Evaluate the log likelihood ratio for presence of a Gaussian spot
    L = 1.0 - stable_divide_array(
        C, n_pixels * Sgc2 * (B - A**2), zero=0.001
    )

    # Set probability of detection close to edges to zero
    hw = w // 2
    L[:hw, :] = 1.0
    L[:, :hw] = 1.0
    L[-hw:, :] = 1.0
    L[:, -hw:] = 1.0

    return threshold_image(
        -(n_pixels / 2.0) * np.log(L), t=t, return_filt=return_filt
    )


def hess_det(I, k=1.0, t=200.0, return_filt=False):
    """
    Use the local Hessian determinant of the image as the
    criterion for detection. The Hessian determinant is related
    to the "spot-ness" of an image and is generally a better
    criterion for detection than the Laplacian alone (as in LoG
    filtering).

    args
    ----
        frame       :   2D ndarray
        k           :   float, Gaussian filtering kernel size
        t           :   float, threshold for detection
        return_filt :   bool

    returns
    -------
        If *return_filt*:
            (
                2D ndarray, filtered image;
                2D ndarray, binary image;
                pandas.DataFrame, the detections
            )
        else:
            pandas.DataFrame, the detections

    """
    # Gaussian filter
    I_filt = ndi.gaussian_filter(I, k)

    def derivative2(im, axis, output, mode, cval):
        return ndi.correlate1d(
            im, [-1, 16, -30, 16, -1], axis, output, mode, cval, 0
        )

    # Discrete second derivative in the y direction
    Lyy = derivative2(I_filt, 0, None, "reflect", 0.0) / 12.0

    # Discrete second derivative in the x direction
    Lxx = derivative2(I_filt, 1, None, "reflect", 0.0) / 12.0

    # Laplacian
    Lxy = (Lyy + Lxx) / 12.0

    # Hessian determinant
    doh = (k**4) * (Lyy * Lxx - Lxy**2)

    return threshold_image(doh, t=t, return_filt=return_filt)


def hess_det_var(I, k=1.0, t=200.0, w0=15, w1=9, return_filt=False):
    """
    Similar to hess_det, this uses the local Hessian determinant
    of the image as the criterion for spot detection. However, it
    normalizes the Hessian by its local variance in a hollow ring
    around each point (see quot.helper.hollow_box_var). This endows
    it with a kind of invariance with respect to the absolute
    intensity of the image, resulting in more consistent threshold
    arguments.

    args
    ----
        frame       :   2D ndarray
        k           :   float, Gaussian filtering kernel size
        t           :   float, threshold for detection
        w0          :   int, width of the box around each point
                        in which to calculate the variance
        w1          :   int, width of the hollow subregion inside
                        each box to exclude from variance calculations
        return_filt :   bool

    returns
    -------
        If *return_filt*:
            (
                2D ndarray, filtered image;
                2D ndarray, binary image;
                pandas.DataFrame, the detections
            )
        else:
            pandas.DataFrame, the detections

    """
    # Gaussian filter
    I_filt = ndi.gaussian_filter(I, k)

    def derivative2(im, axis, output, mode, cval):
        return ndi.correlate1d(
            im, [-1, 16, -30, 16, -1], axis, output, mode, cval, 0
        )

    # Discrete second derivative in the y direction
    Lyy = derivative2(I_filt, 0, None, "reflect", 0.0) / 12.0

    # Discrete second derivative in the x direction
    Lxx = derivative2(I_filt, 1, None, "reflect", 0.0) / 12.0

    # Laplacian
    Lxy = (Lyy + Lxx) / 12.0

    # Hessian determinant
    doh = (k**4) * (Lyy * Lxx - Lxy**2)

    # Local variance in the Hessian determinant
    doh_var = hollow_box_var(doh, w0=w0, w1=w1)
    doh = doh / np.sqrt(doh_var)

    return threshold_image(doh, t=t, return_filt=return_filt)


def hess_det_broad_var(I, k=1.0, t=200.0, w0=15, w1=9, return_filt=False):
    """
    Use the local Hessian determinant of the image as the
    criterion for detection. This method uses a broader definition
    of the Laplacian kernel than hess_det() or hess_det_var().

    This method also normalizes the Hessian determinant against
    its local variance to give it the property of intensity
    invariance. Otherwise, the broader Laplacian kernel tends
    to produce quite different threshold values for different
    cameras.

    args
    ----
        frame       :   2D ndarray
        k           :   float, Gaussian filtering kernel size
        t           :   float, threshold for detection
        w0          :   int, width of the box around each point
                        in which to calculate the variance
        w1          :   int, width of the hollow subregion inside
                        each box to exclude from variance calculations
        return_filt :   bool

    returns
    -------
        If *return_filt*:
            (
                2D ndarray, filtered image;
                2D ndarray, binary image;
                pandas.DataFrame, the detections
            )
        else:
            pandas.DataFrame, the detections

    """
    # Gaussian filter
    I_filt = ndi.gaussian_filter(I, k)

    def derivative2(im, axis, output, mode, cval):
        return ndi.correlate1d(
            im, [9, 28, -6, -62, -6, 28, 9], axis, output, mode, cval, 0
        )

    # Discrete second derivative in the y direction
    Lyy = derivative2(I_filt, 0, None, "reflect", 0.0) / 12.0

    # Discrete second derivative in the x direction
    Lxx = derivative2(I_filt, 1, None, "reflect", 0.0) / 12.0

    # Laplacian
    Lxy = (Lyy + Lxx) / 12.0

    # Hessian determinant
    doh = (k**4) * (Lyy * Lxx - Lxy**2)

    # Normalize by the local variance in the Hessian determinant
    doh_var = hollow_box_var(doh, w0=w0, w1=w1)
    doh = doh / np.sqrt(doh_var)

    return threshold_image(doh, t=t, return_filt=return_filt)


def llr_rect(I, w0=3, w1=11, t=20.0, return_filt=False):
    """
    Perform a log-likelihood ratio test for the presence of
    square spots of size *w0* in an image using a test equivalent
    to llr(). While squares only roughly approximate real spots,
    the test can be performed extremely fast due to the fact that
    only uniform filtering is required.

    args
    ----
        I           :   2D ndarray
        w0          :   int, spot kernel size
        w1          :   int, background kernel size
        t           :   float, threshold for spot detection
        return_filt :   bool, also return filtered image

    returns
    -------
        if return_filt:
        (
            2D ndarray, the post-convolution image;
            2D ndarray, the thresholded binary image;
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot
        )
        else
            2D ndarray, shape (n_spots, 2), the y and x
                coordinates of each spot

    """
    n_pixels = w1**2

    # Compute normalization factor
    Suc2 = n_pixels * ((1.0 / (w0**2)) - (1.0 / (w1**2)))

    # Perform convolutions for detection
    A = ndi.uniform_filter(I, w1)
    C = (ndi.uniform_filter(I, w0) - A) ** 2
    B = ndi.uniform_filter(I**2, w1)

    # Calculate likelihood of squares centered on each pixel
    L = 1.0 - stable_divide_array(C, Suc2 * (B - (A**2)), zero=0.0001)

    # Set probability of detection close to edges to zero
    hw = w1 // 2
    L[:hw, :] = 1.0
    L[:, :hw] = 1.0
    L[-hw:, :] = 1.0
    L[:, -hw:] = 1.0

    # Take log-likelihood and threshold to find spots
    return threshold_image(
        -(n_pixels / 2.0) * np.log(L), t=t, return_filt=return_filt
    )


#############################
## MAIN DETECTION FUNCTION ##
#############################

# All available detection methods
METHODS = {
    #'gauss': gauss,
    #'centered_gauss': centered_gauss,
    #'mle_amp': mle_amp,
    "dog": dog,
    "log": log,
    #'dou': dou,
    #'min_max': min_max,
    #'gauss_filt_min_max': gauss_filt_min_max,
    "llr": llr,
    #'hess_det': hess_det,
    #'hess_det_var': hess_det_var,
    #'hess_det_broad_var': hess_det_broad_var,
    "llr_rect": llr_rect,
}

t_default = {
    "dog": 200.0,
    "log": 200.0,
    "llr": 20.0,
    "llr_rect": 20.0,
}


def detect(I, method=None, **kwargs):
    """
    Run spot detection on a single image according to a
    detection method.

    args
    ----
        I       :   2D ndarray (YX), image frame
        method  :   str, a method name in METHODS
        kwargs  :   special argument to the method

    returns
    -------
        2D ndarray of shape (n_spots, 2), the Y and
            X coordinates of identified spots

    """
    # Get the desired method
    method_f = METHODS.get(method)
    if method_f is None:
        raise KeyError(
            "Method %s not available; available "
            "methods are %s" % (method, ", ".join(METHODS.keys()))
        )

    # Enforce float64, required for some methods
    I = I.astype(np.float64)
    if "t" in kwargs:
        kwargs["t"] *= t_default.get(method)

    # Run detection
    return method_f(I, **kwargs)



#!/usr/bin/env python
"""
mask.py -- apply binary masks to an SPT movie

"""
import os
import sys
import numpy as np 
import pandas as pd 
import matplotlib.pyplot as plt 
from matplotlib.path import Path 
from scipy import interpolate 
from scipy.spatial import distance_matrix 
from tqdm import tqdm 

class MaskInterpolator(object):
    """
    Given a set of 2D masks, interpolate the masks between frames to 
    generate an approximation to the mask for intermediate frames.

    In more detail:

    We have a set of 2D masks defined in frames F0, F1, ..., Fn. Masks
    are defined as a ringlike, connected set of points. We wish to use
    the masks defined in frames F(i) and F(i+1) to estimate the shape
    of the mask in any intermediate frame, assuming the mask varies 
    smoothly/linearly between frame F(i) and F(i+1).

    Instantiation of the LinearMaskInterpolator() accomplishes this, 
    using either linear or spline interpolation. The resulting object
    can then be passed any frame index between F0 and Fn and will 
    generate the corresponding 2D mask.

    The resulting object can be passed a frame index, and will generate
    the corresponding 2D mask.

    init
    ----
        mask_edges      :   list of 2D ndarray of shape (n_points, 2),
                            the Y and X coordinates for each mask at 
                            each frame
        mask_frames     :   list of int, the frame indices corresponding
                            to each mask edge
        n_vertices      :   int, the number of vertices to use per 
                            interpolated mask
        interp          :   str, "linear" or "cubic", the type of 
                            interpolation to use. Note that at least 4 
                            masks are required for cubic spline interpolation.
        plot            :   bool, show a plot of the vertex matching between
                            interpolated masks during initialization, for QC

    methods
    -------
        __call__        :   determine whether each of a set of points lies
                            inside or outside the mask
        interpolate     :   generate an interpolated mask edge for an arbitrary
                            frame lying between the minimum and maximum frames
                            for this object

    """
    def __init__(self, mask_edges, mask_frames, n_vertices=101, 
        interp_kind="linear", plot=True):

        assert len(mask_edges) == len(mask_frames)
        assert interp_kind in ["linear", "cubic"]

        self.mask_edges = mask_edges 
        self.mask_frames = mask_frames 
        self.interp_kind = interp_kind
        self.n_frames = len(self.mask_frames)
        self.n_vertices = n_vertices 
        self.plot = plot 

        # If passed only a single frame, then the interpolator
        # always returns a simple static 2D shape
        self.static = (self.n_frames == 1)

        # Otherwise, generate interpolator objects to reconstruct
        # the mask for any other frame
        if not self.static:
            self._generate_mask_matches()
            self._generate_interpolators()

    def __call__(self, points, frame_indices, progress_bar=False):
        """
        Given a set of points, determine whether each point lies inside
        or outside the present mask.

        args
        ----
            points          :   2D ndarray of shape (n_points, 2), the 
                                YX coordinates for each point
            frame_indices   :   1D ndarray of shape (n_points,), the 
                                frame indices corresponding to each point
            progress_bar    :   bool, show a progress bar

        returns
        -------
            1D ndarray of shape (n_points,), dtype bool

        """
        assert points.shape[0] == frame_indices.shape[0]

        # Format as ndarray
        if isinstance(points, pd.DataFrame):
            points = np.asarray(points)
        if isinstance(frame_indices, pd.Series):
            frame_indices = np.asarray(frame_indices)

        unique_frames = np.unique(frame_indices)
        if progress_bar:
            unique_frames = tqdm(unique_frames)
        assignments = np.empty(points.shape[0])
        for frame_index in unique_frames:
            mask = Path(self.interpolate(frame_index), closed=True)
            in_frame = frame_indices == frame_index 
            assignments[in_frame] = mask.contains_points(points[in_frame,:])
        return assignments

    def interpolate(self, frame):
        """
        Interpolate the mask edges for a given frame index.

        args
        ----
            frame       :   int, the frame index

        returns
        -------
            2D ndarray of shape (self.n_vertices, 2), the YX
                coordinates for the points along the edge of 
                the mask

        """
        if self.static:
            return self.mask_edges[0]
        else:

            # Frame lies outside of interpolable range
            if not self._within_interpolation_range(frame):
                return self.mask_edges[:,-1,:]

            # Otherwise interpolate the mask
            result = np.empty((self.n_vertices, 2), dtype=np.float64)
            result[:,0] = [self.y_interpolators[i](frame) for i in range(self.n_vertices)]
            result[:,1] = [self.x_interpolators[i](frame) for i in range(self.n_vertices)]

            return result 

    def _within_interpolation_range(self, frame):
        """
        Return True if the frame lies inside the interpolation
        range for this MaskInterpolator instance.

        """
        return (frame >= self.mask_frames[0]) and (frame <= self.mask_frames[-1])

    def _generate_mask_matches(self):
        """
        Given the set of points that defines the edge of each mask
        used to instantiate this object, upsample the masks to the 
        same number of points and match each point with the points
        in other masks for subsequent interpolation.

        Resets the self.mask_edges attribute to the result.

        """
        # The total number of sets of mask edges to match 
        n_masks = len(self.mask_edges)

        # The final set of matched points, appropriate for 
        # interpolation
        result = np.zeros((self.n_vertices, n_masks, 2), dtype=np.float64)

        # Upsample each of the masks to the same number of points
        for j in range(n_masks):
            self.mask_edges[j] = upsample_2d_path(self.mask_edges[j],
                kind="cubic", n_vertices=self.n_vertices)

        result[:,0,:] = self.mask_edges[0]

        # For each sequential combination of masks, match the
        # interpolated mask points 
        for j in range(1, n_masks):
            result[:,j,:] = match_vertices(
                result[:,j-1,:],
                self.mask_edges[j],
                method="global",
                plot=self.plot
            )

        self.mask_edges = result 

    def _generate_interpolators(self):
        """
        For each point that defines the edge of this mask, generate Y
        and X interpolators that enable a mask to be reconstructed
        for a given image frame.

        """
        self.y_interpolators = []
        self.x_interpolators = []

        for vertex in range(self.n_vertices):

            # Generate the interpolator object for the y-index 
            I = interpolate.interp1d(self.mask_frames, self.mask_edges[vertex, :, 0],
                kind=self.interp_kind)
            self.y_interpolators.append(I)

            # Generate the interpolator object for the y-index 
            I = interpolate.interp1d(self.mask_frames, self.mask_edges[vertex, :, 1],
                kind=self.interp_kind)
            self.x_interpolators.append(I)

def upsample_2d_path(points, kind="cubic", n_vertices=101):
    """
    Upsample a 2D path by interpolation.

    args
    ----
        points      :   2D ndarray of shape (n_points, 2), the 
                        Y and X coordinates of each point in the 
                        path, organized sequentially
        kind        :   str, the kind of spline interpolation
        n_vertices  :   int, the number of points to use in the 
                        upsampled path

    returns
    -------
        2D ndarray of shape (n_vertices, 2), the upsampled 
            path

    """
    P = np.concatenate((points, np.array([points[0,:]])), axis=0)
    t = np.arange(P.shape[0])
    fy = interpolate.interp1d(t, P[:,0], kind=kind)
    fx = interpolate.interp1d(t, P[:,1], kind=kind)
    result = np.empty((n_vertices, 2), dtype=np.float64)
    new_t = np.linspace(0, t.max(), n_vertices)
    result[:,0] = fy(new_t)
    result[:,1] = fx(new_t)
    return result 

def shoelace(points):
    """
    Shoelace algorithm for computing the oriented area of a 2D
    polygon. This area is positive when the points that define
    the polygon are arranged counterclockwise, and negative
    otherwise.

    args
    ----
        points      :   2D ndarray, shape (n_points, 2), the 
                        vertices of the polygon

    returns
    -------
        float, the oriented volume of the polygon defined by 
            *points*

    """
    return ((points[1:,0] - points[:-1,0]) * \
        (points[1:,1] + points[:-1,1])).sum()

def circshift(points, shift):
    """
    Circularly shift a set of points.

    args
    ----
        points      :   ndarray of shape (n_points, D), the 
                        D-dimensional coordinates of each point 
        shift       :   int, the index of the new starting point

    returns
    -------
        2D ndarray of shape (n_points, D), the same points
            but circularly shifted

    example
    -------
        points_before = np.array([
            [1, 2],
            [3, 4],
            [5, 6]
        ])

        points_after = circshift(points_before, 1)

        points_after -> np.array([
            [3, 4],
            [5, 6],
            [1, 2]
        ])

    """
    out = np.empty(points.shape, dtype=points.dtype)
    n = out.shape[0]
    shift = shift % n 
    if len(points.shape) == 1:
        out[:n-shift] = points[shift:]
        out[n-shift:] = points[:shift]
    elif len(points.shape) == 2:
        out[:n-shift,:] = points[shift:,:]
        out[n-shift:,:] = points[:shift,:]
    return out 

def match_vertices(vertices_0, vertices_1, method="closest", plot=False):
    """
    Given two polygons with the same number of vertices, match
    each vertex in the first polygon with the "closest" vertex
    in the second polygon.

    "closest" is in quotation marks here because, before matching,
    we align the two polygons by their mean position, so that the
    same match is returned regardless of whole-polygon shifts.

    args
    ----
        vertices_0      :   2D ndarray, shape (n_points, 2), the 
                            YX coordinates for the vertices of the 
                            first polygon
        vertices_1      :   2D ndarray, shape (n_points, 2), the 
                            YX coordinates for the vertices of the 
                            second polygon
        method          :   str, the method to use to match vertices.
                            "closest": use the closest point between
                            the two masks as the anchor point. 
                            "global": use the permutation that minimizes
                            the total distance between the two sets
                            of vertices.
        plot            :   bool, show the result

    returns
    -------
        2D ndarray, shape (n_points, 2), the vertices of the 
            second polygon circularly permuted to line them up
            with the matching vertex in the first polygon

    """
    assert vertices_0.shape == vertices_1.shape
    assert method in ["closest", "global"], "method must be either 'closest' or 'global'"

    # The final assignments
    indices_1 = np.arange(vertices_1.shape[0])

    # Deal only with the positions of each vertex relative to the 
    # respective mean
    P0 = vertices_0 - vertices_0.mean(axis=0)
    P1 = vertices_1 - vertices_1.mean(axis=0)

    # Make sure the points both proceed in the same direction (CW or CCW)
    if shoelace(P0) * shoelace(P1) < 0:
        P1 = P1[::-1,:]
        indices_1 = indices_1[::-1]

    # Align masks by simply looking for two vertices that are closest
    if method == "closest":
        distances = distance_matrix(P0, P1)
        m = np.argmin(distances.ravel())
        y, x = m // P0.shape[0], m % P0.shape[0]
        shift = (x - y) % P0.shape[0]

    # Align masks by minimizing the sum of the distances between all
    # vertices for all possible matches
    elif method == "global":
        curr = 0
        ss = np.inf 
        for x in range(P1.shape[0]):
            shift_P1 = circshift(P1, x)
            tot_dist = ((P0 - shift_P1)**2).sum()
            if tot_dist < ss:
                curr = x 
                ss = tot_dist 
        shift = curr 

    # Align the masks
    indices_1 = circshift(indices_1, shift)
    vertices_1 = vertices_1[indices_1, :]

    # Show the resulting set of vertex matches, if desired
    if plot:
        fig, ax = plt.subplots(figsize=(3, 3))
        ax.scatter(vertices_0[:,0], vertices_0[:,1], cmap="viridis",
            c=np.arange(vertices_0.shape[0]))
        ax.scatter(vertices_1[:,0], vertices_1[:,1], cmap="viridis",
            c=np.arange(vertices_1.shape[0]))
        for j in range(vertices_0.shape[0]):
            ax.plot([vertices_0[j,0], vertices_1[j,0]], [vertices_0[j,1], vertices_1[j,1]], 
                color="k", linestyle='-')
        ax.set_aspect('equal')
        ax.set_title("Mask alignment")
        plt.show()

    return vertices_1 




import numpy as np
import pandas as pd
import dask.array as da
from dask_image.ndfilters import convolve


def smooth_average(x, w):
    r = (da.arange(w) - w / 2) / (w / 6)
    gaussian = da.exp(-0.5 * r**2)
    gaussian /= da.sum(gaussian)
    return convolve(x, gaussian, "reflect").compute()


def mean_intensity_center(data):
    T, W, H = data.shape
    # indices = tuple(
    #    np.ogrid[0:T, (W // 2 - 10) : (W // 2 + 10), (H // 2 - 10) : (H // 2 + 10)]
    # )
    # center = data[indices]
    # print(indices)
    center = data[:, (W // 2 - 10) : (W // 2 + 10), :]
    center = center[:, :, (H // 2 - 10) : (H // 2 + 10)]
    average = center.mean(axis=(1, 2))
    smoothed_average = smooth_average(average, min(500, average.shape[0]))
    return pd.DataFrame.from_dict(
        {
            "intensity": smoothed_average,
            "frame": np.arange(smoothed_average.shape[0]),
        }
    )



#!/usr/bin/env python
"""
__init__.py

ALL THE CODE IN THIS FOLDER WAS COPIED/ADAPTED FROM 
https://github.com/alecheckert/quot
Which is under MIT LICENSE

"""
# Enforce PySide2 backend to pyqtgraph
# import PySide2
# import pyqtgraph

# Core functions to run localization and tracking on single files
# or directories
# from .core import (
#    localize_file,
#    track_file,
#    track_directory,
#    retrack_file,
#    retrack_files,
# )

# Read and filter image files
#from .read import ImageReader, read_config
#from .chunkFilter import ChunkFilter

# Find spots
from .findSpots import detect

# Localize spots to subpixel resolution
from .subpixel import localize, localize_frame

# Reconnection spots into trajectories
from .track import track



from .._widget import PalmariWidget
from ..data_structure.acquisition import Acquisition
from ..data_structure.experiment import Experiment
from ..processing.image_pipeline import ImagePipeline
from glob import glob
import numpy as np

# make_napari_viewer is a pytest fixture that returns a napari viewer object
# capsys is a pytest fixture that captures stdout and stderr output streams
def test_example_q_widget(make_napari_viewer, capsys):
    # make viewer and add an image layer using our fixture
    viewer = make_napari_viewer()
    viewer.add_image(np.random.random((100, 100)))

    # create our widget, passing in the viewer
    my_widget = PalmariWidget(viewer)

    # read captured output and check that it's as we expected
    # captured = capsys.readouterr()
    # assert captured.out == "napari has 1 layers\n"
    assert True


#!/usr/bin/env python
"""
__main__.py -- launch batch analyses from the command line

"""
import os
import click
from shutil import copyfile

from .read import read_config, save_config
from .core import track_directory, track_file

ACCEPTABLE_EXTS = [".nd2", ".tif", ".tiff"]

@click.command()
@click.argument("target_dir", type=str)
@click.argument("config_path", type=str)
@click.option("-e", "--ext", type=str, default=".nd2",
    help="default .nd2")
@click.option("-n", "--n_threads", type=int, default=1,
    help="default 1")
@click.option("-o", "--out_dir", type=str, default=None,
    help="default same as image files")
@click.option("-c", "--contains", type=str, default=None,
    help="default None")
def batch_track(target_dir, config_path, ext, n_threads,
    out_dir, contains):
    """
    Run batch detection and tracking on all image files in 
    a target directory.

    TARGET_DIR  :  a directory with image files

    CONFIG_PATH :  path to a TOML configuration file

    EXT         :  the extension to look for (eg .nd2, .tif)

    N_THREADS   :  the number of parallel threads to use

    OUT_DIR     :  directory for output CSV files

    CONTAINS    :  a substring that all target files must contain

    """
    # Load the experimental configuration
    try:
        config = read_config(config_path)
    except:
        raise RuntimeError("Could not load config file {}".format(config_path))

    # If passed an image file instead of a directory, just run tracking
    # on this file
    if os.path.isfile(target_dir) and (os.path.splitext(target_dir)[-1] in ACCEPTABLE_EXTS):

        # Specify output file
        if (not out_dir is None) and (".csv" in out_dir):
            out_csv = out_dir 
        elif (not out_dir is None):
            if not os.path.isdir(out_dir):
                os.mkdir(out_dir)
            out_csv = os.path.join(
                out_dir,
                "{}_trajs.csv".format(
                    os.path.splitext(os.path.basename(target_dir))[0]
                )
            )
        else:
            out_csv = "{}_trajs.csv".format(os.path.splitext(target_dir)[0])

        # Run tracking
        track_file(
            target_dir,
            out_csv=out_csv,
            progress_bar=True,
            **config
        )

    # Otherwise run tracking on matching image files in the passed
    # directory
    elif os.path.isdir(target_dir):
        track_directory(
            target_dir,
            ext=ext,
            num_workers=n_threads,
            out_dir=out_dir,
            contains=contains,
            **config
        )

    else:
        raise RuntimeError("No file or directory: {}".format(target_dir))

@click.command()
@click.argument("config_path", type=str)
def make_naive_config(config_path):
    """
    Save a naive config file with some stock settings to a target
    path. 

    """
    src = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "stock_config.toml"
    )
    copyfile(src, config_path)

if __name__ == "__main__":
    batch_track()



#!/usr/bin/env python
"""
track.py -- reconnect localizations into trajectories

"""
# Numeric
import numpy as np

# Dataframes
import pandas as pd

# Distance between two sets of 2D points
from scipy.spatial import distance_matrix

# Hungarian algorithm
from munkres import Munkres
from scipy.optimize import linear_sum_assignment

# Custom utilities
from .helper import connected_components

# Deep copy
from copy import copy

# Filter on the number of spots per frame
from .trajUtils import filter_on_spots_per_frame
from tqdm import tqdm

##################################
## LOW-LEVEL TRACKING UTILITIES ##
##################################


class ScipyMunkres:
    def compute(self, W):
        rows, cols = linear_sum_assignment(W)
        return [(i, j) for i, j in zip(rows, cols)]


# hungarian_solver = Munkres()
hungarian_solver = ScipyMunkres()


def is_duplicates(trajs):
    """
    Return True if there are duplicates in a set of Trajectories.

    """
    if len(trajs) < 2:
        return False
    for j in range(len(trajs) - 1):
        for i in range(j + 1, len(trajs)):
            R = trajs[i].get_slice()[:, :2] == trajs[j].get_slice()[:, :2]
            if isinstance(R, bool):
                if R:
                    return True
            elif R.all():
                return True
            else:
                pass
    return False


def has_match(trajs_0, trajs_1):
    """
    Return True if a trajectory in trajs_0 exactly matches a trajectory
    in trajs_1.

    """
    for i in range(len(trajs_0)):
        for j in range(len(trajs_1)):
            R = trajs_0[i].get_slice()[:, :2] == trajs_1[j].get_slice()[:, :2]
            if isinstance(R, bool):
                if R:
                    return True
            elif R.all():
                return True
            else:
                pass
    return False


class Trajectory:
    """
    Convenience class used internally by track_locs().

    A Trajectory object specifies a set of indices to
    localizations in a large array that are to be reconnected
    into trajectories.

    It also holds a reference to the original array, so
    that it can grab information about its localization when
    necessary.

    When Trajectories are not reconnected in a given frame,
    their blink counter (self.n_blinks) is incremented. When
    this exceeds *max_blinks*, the Trajectories are marked
    for termination.

    The Trajectory class is also convenient to hold associated
    information about the tracking problem, such as the
    number of competing trajectories and localizations, etc.,
    that are returned at the end of tracking.

    init
    ----
        start_idx       :   int, the index of the first
                            localization in this Trajectory
        locs            :   2D ndarray, all localizations
        subproblem_shape:   (int, int), the number of trajs
                            and locs in the subproblem that
                            created this trajectory
        max_blinks      :   int, the maximum tolerated number
                            of gaps in tracking

    """

    def __init__(self, start_idx, locs, subproblem_shape, max_blinks=0):
        self.indices = [int(start_idx)]
        self.locs = locs
        self.max_blinks = max_blinks
        self.n_blinks = 0
        self.active = True
        self.subproblem_shapes = [subproblem_shape]

    def add_index(self, idx, subproblem_shape):
        """
        Extend this trajectory by one localization.

        args
        ----
            idx     :   int, the index of the localization
                        in self.locs to add
            subproblem_shape    :   (int, int), the size
                        of the tracking subproblem in which
                        this localization was added

        """
        self.indices.append(int(idx))
        self.subproblem_shapes.append(subproblem_shape)

    def blink(self):
        """
        Skip a frame. If a Trajectory has been in blink for
        more than *self.n_blinks* frames, it marks itself
        for termination by setting self.active = False.

        """
        self.n_blinks += 1
        if self.n_blinks > self.max_blinks:
            self.active = False

    def get_slice(self):
        """
        Return the slice of the localization array that
        corresponds to this trajectory.

        returns
        -------
            2D ndarray of shape (traj_len, 5)

        """
        return self.locs[tuple(self.indices), :]

    def last_pos(self):
        """
        Return the last known position of this Trajectory.

        returns
        -------
            (float y, float x), in pixels

        """
        return self.locs[self.indices[-1], 2:4]


def traj_loc_distance(trajs, locs):
    """
    Return the distance between each trajectory and each
    localization.

    args
    ----
        trajs       :   list of Trajectory
        locs        :   2D ndarray with columns loc_idx,
                        frame, y, x, I0

    returns
    -------
        2D ndarray D, where D[i,j] is the distance between
            Trajectory i and localization j

    """
    return distance_matrix(
        np.asarray([t.last_pos() for t in trajs]), locs[:, 2:4]
    )


def diffusion_weight_matrix(
    trajs,
    locs,
    frame_interval=0.00548,
    pixel_size_um=0.16,
    k_return_from_blink=1.0,
    d_max=5.0,
    y_diff=0.9,
    search_radius=2.5,
    d_bound_naive=0.1,
    init_cost=50.0,
):
    """
    Generate the weight matrix for reconnection between a set of
    Trajectories and a set of localizations for the "diffusion"
    method.

    In this method, the weight of reconnecting trajectory A to
    localization B is equal to the negative log likelihood of
    trajectory A diffusing to localization B in the relevant
    frame interval (equal to *frame_interval* if there are no
    gaps). The likelihood is evaluated with a 2D Brownian motion
    model.

    A weighted combination of two such negative log-likelihoods
    is used. The first assumes a diffusion coefficient equal to
    the maximum likelihood diffusion coefficient for that
    trajectory (using the MSD method). The second assumes a
    diffusion coefficient equal to d_max. The relative weights
    of the two estimates are set by *y_diff*.

    args
    ----
        trajs           :   list of Trajectory
        locs            :   2D ndarray, localizations to consider
                            for connection
        frame_interval  :   float, seconds
        pixel_size_um   :   float, um
        k_return_from_blink :   float, penalty to return a trajectory
                                from blinking status
        d_max           :   float, the maximum expected diffusion
                            coefficient in um^2 s^-1
        y_diff          :   float, the relative influence of the
                            particle's local history on its estimated
                            diffusion coefficient
        search_radius   :   float, um
        d_bound_naive   :   float, naive estimate for a particle's
                            local diffusion coefficient, um^2 s^-1
        init_cost       :   float, static cost to initialize a new
                            trajectory when reconnections are available
                            in the search radius

    returns
    -------
        2D ndarray of shape (n_trajs, n_locs), the weights
            for reconnection

    """
    n_traj = len(trajs)
    n_locs = locs.shape[0]
    n_dim = n_traj + n_locs

    # Unit conversions
    search_radius_pxl = search_radius / pixel_size_um
    max_var2 = 2.0 * d_max * frame_interval / (pixel_size_um**2)

    # The weight matrix
    W = np.zeros((n_dim, n_dim), dtype="float64")

    # Set the static cost of starting a new trajectory
    for li in range(n_locs):
        W[n_traj:, li] = init_cost
    W[:, n_locs:] = init_cost

    # For each trajectory, calculating the weight of
    # assignments to each localization
    for ti in range(n_traj):

        # Last known position of this trajectory
        last_pos = trajs[ti].last_pos()

        # Penalize blinking trajectories
        L_blink = -k_return_from_blink * trajs[ti].n_blinks + np.log(
            k_return_from_blink
        )

        # Distances to each localization
        R = traj_loc_distance([trajs[ti]], locs)[0, :]
        R2 = R**2

        # Estimate the local diffusion coefficient of this
        # trajectory.

        # If no prior displacements are available, use the
        # naive estimate
        if len(trajs[ti].indices) == 1:
            local_var2 = (
                2 * d_bound_naive * frame_interval * (1 + trajs[ti].n_blinks)
            )

        # Otherwise, use the MSD method
        else:
            traj_slice = trajs[ti].get_slice()
            frames = traj_slice[:, 1].astype("int64")
            pos = traj_slice[:, 2:4]
            delta_frames = frames[1:] - frames[:-1]
            local_var2 = (
                0.5
                * (
                    ((pos[1:, :] - pos[:-1, :]) ** 2).sum(1) / delta_frames
                ).mean()
                * (1 + trajs[ti].n_blinks)
            )

        # Log-likelihood of diffusing from last
        # known position to each new position
        diff = y_diff * (R / local_var2) * np.exp(-R2 / (2 * local_var2)) + (
            1 - y_diff
        ) * (R / max_var2) * np.exp(-R2 / (2 * max_var2))
        L_diff = np.log(diff)
        L_diff[diff <= 0] = -np.inf

        # Make sure we do NOT reconnect trajectories to localizations
        # outside of their search radii
        L_diff[R > search_radius_pxl] = -np.inf

        # Assign reconnection weight
        W[ti, :n_locs] = -(L_blink + L_diff)

    return W


def euclidean_weight_matrix(
    trajs,
    locs,
    pixel_size_um=0.16,
    scale=1.0,
    search_radius=2.5,
    init_cost=50.0,
    **kwargs
):
    """
    Generate the weight matrix for reconnection between
    Trajectories and localizations for the "euclidean"
    reconnection method.

    Here, the weight to reconnect traj I with localization J
    is just the distance between the last known position of I
    and J, scaled by the constant *scale*.

    If J is outside the search radius of I, the weight is
    infinite.

    The weight to drop *I* or to start a new trajectory from
    *J* when other reconnections are available is *init_cost*.

    args
    ----
        trajs           :   list of Trajectory
        locs            :   2D ndarray, localizations to consider
                            for connection
        pixel_size_um   :   float, um
        scale           :   float, inflation factor for the distances
        search_radius   :   float, um
        init_cost       :   float, penalty for not performing
                            available reconnections
        kwargs          :   discarded

    returns
    -------
        2D ndarray of shape (n_trajs, n_locs), the reconnection
            weights

    """
    n_traj = len(trajs)
    n_locs = locs.shape[0]
    n_dim = n_traj + n_locs

    # Weight matrix
    W = np.zeros((n_dim, n_dim), dtype="float64")

    # Distance from each trajectory to each localization
    distances = traj_loc_distance(trajs, locs)

    # Reconnections out of the search radius are impossible
    out_of_radius = distances > search_radius / pixel_size_um
    distances[out_of_radius] = np.inf

    # Rescale if desired
    if scale != 1.0:
        distances[~out_of_radius] = distances[~out_of_radius] * scale

    # Weight of traj:loc reconnection is proportional to Euclidean distance
    W[:n_traj, :n_locs] = distances

    # Static cost to start new trajectory
    W[n_traj:, :] = init_cost

    # Penalize reconnecting to nothing
    W[:n_traj, n_locs:] = init_cost

    return W


######################################
## FRAME-FRAME RECONNECTION METHODS ##
######################################


def reconnect_conservative(
    trajs,
    locs,
    locs_array,
    max_blinks=0,
    frame_interval=0.00548,
    pixel_size_um=0.16,
    **kwargs
):
    """
    Only reassign trajs to locs when the assignment is
    unambiguous (1 traj, 1 loc within the search radius).

    For all other trajectories, terminate them.
    For all other locs, start new trajectories.

    args
    ----
        trajs           :   list of Trajectory
        locs            :   2D ndarray with columns loc_idx, frame,
                            y, x, I0
        max_blinks      :   int
        frame_interval  :   float
        pixel_size_um   :   float
        kwargs          :   ignored, possibly passed due to upstream
                            method disambiguation

    returns
    -------
        list of Trajectory

    """
    out = []

    # A single localization and trajectory pair - assignment
    # is unambiguous. This is the only situation where reconnection
    # is allowed.
    n_trajs = len(trajs)
    n_locs = locs.shape[0]
    if n_trajs == 1 and n_locs == 1:
        trajs[0].add_index(locs[0, 0], (n_trajs, n_locs))
        out = trajs

    # Multiple localizations and/or trajectories
    else:

        # Terminate all existing trajectories
        for ti in range(n_trajs):
            trajs[ti].active = False
        out += trajs

        # Start new trajectories from all localizatoins
        for li in range(n_locs):
            out.append(
                Trajectory(
                    locs[li, 0],
                    locs_array,
                    (n_trajs, n_locs),
                    max_blinks=max_blinks,
                )
            )

    return out


WEIGHT_MATRIX_METHODS = {
    "diffusion": diffusion_weight_matrix,
    "euclidean": euclidean_weight_matrix,
}


def reconnect_hungarian(
    trajs,
    locs,
    locs_array,
    max_blinks=0,
    weight_method=None,
    min_I0=0.0,
    **kwargs
):
    """
    Assign Trajectories to localizations by assigning each
    possible reconnection a weight, then finding the assignment
    that minimizes the summed weights with the Hungarian
    algorithm.

    args
    ----
        trajs           :   list of Trajectory
        locs            :   2D ndarray, localizations to consider
                            for connection
        locs_array      :   2D ndarray, all localizations in this
                            movie
        max_blinks      :   int
        weight_method   :   str, the method to use to generate
                            the weight matrix
        min_I0          :   float, minimum intensity to start
                            a new Trajectory
        kwargs          :   to weight_method

    returns
    -------
        list of Trajectory

    """
    out = []

    # Get the size of the assignment problem
    n_traj = len(trajs)
    n_locs = locs.shape[0]

    # Unambiguous - only one trajectory and localization
    if n_traj == 1 and n_locs == 1:
        trajs[0].add_index(locs[0, 0], (n_traj, n_locs))

    # Otherwise, solve the assignment problem by finding
    # weights and minimizing with the Hungarian algorithm
    else:

        # Get the weight matrix for reconnection
        W = WEIGHT_MATRIX_METHODS[weight_method](trajs, locs, **kwargs)

        # Minimize negative log likelihood with
        # Hungarian algorithm
        for i, j in hungarian_solver.compute(W):

            # traj:loc
            if (i < n_traj) and (j < n_locs):
                trajs[i].add_index(locs[j, 0], (n_traj, n_locs))

            # traj:(empty)
            elif (i < n_traj) and (j >= n_locs):
                trajs[i].blink()

            # (empty):loc
            elif (j < n_locs) and (i >= n_traj):
                if locs[j, 4] >= min_I0:
                    out.append(
                        Trajectory(
                            locs[j, 0],
                            locs_array,
                            (n_traj, n_locs),
                            max_blinks=max_blinks,
                        )
                    )
            else:
                pass

    # Combine new and existing trajs
    out += trajs
    return out


def reconnect_diffusion(
    trajs, locs, locs_array, max_blinks=0, min_I0=0.0, **kwargs
):
    """
    Assign Trajectories to localizations on the basis of their
    expected probability of diffusion and their blinking status.

    Each of the Trajectories is assumed to be a Brownian motion
    in 2D. Its diffusion coefficient is evaluated from its
    history by MSD if it is greater than length 1, or from
    d_bound_naive otherwise.

    args
    ----
        trajs           :   list of Trajectory
        locs            :   2D ndarray, localizations to consider
                            for connection
        locs_array      :   2D ndarray, all localizations in this
                            movie
        max_blinks      :   int
        frame_interval  :   float, seconds
        pixel_size_um   :   float, um
        min_I0          :   float, AU
        k_return_from_blink :   float, penalty to return a trajectory
                                from blinking status
        d_max           :   float, the maximum expected diffusion
                            coefficient in um^2 s^-1
        y_diff          :   float, the relative influence of the
                            particle's local history on its estimated
                            diffusion coefficient
        search_radius   :   float, um
        d_bound_naive   :   float, naive estimate for a particle's
                            local diffusion coefficient, um^2 s^-1

    returns
    -------
        list of Trajectory

    """
    return reconnect_hungarian(
        trajs,
        locs,
        locs_array,
        weight_method="diffusion",
        max_blinks=0,
        min_I0=min_I0,
        **kwargs
    )


def reconnect_euclidean(
    trajs, locs, locs_array, max_blinks=0, min_I0=0.0, **kwargs
):
    """
    Assign Trajectories to localizations purely by minimizing
    the total Trajectory-localization distances.

    args
    ----
        trajs           :   list of Trajectory
        locs            :   2D ndarray, localizations to consider
                            for connection
        locs_array      :   2D ndarray, all localizations in this
                            movie
        max_blinks      :   int
        min_I0          :   float, minimum intensity to start a
                            new trajectory
        pixel_size_um   :   float, um
        scale           :   float, inflation factor for the distances
        search_radius   :   float, um
        init_cost       :   float, cost to start a new trajectory
                            if reconnections are available

    returns
    -------
        list of Trajectory

    """
    return reconnect_hungarian(
        trajs,
        locs,
        locs_array,
        weight_method="euclidean",
        max_blinks=0,
        min_I0=min_I0,
        **kwargs
    )


########################################
## ALL AVAILABLE RECONNECTION METHODS ##
########################################

METHODS = {
    "conservative": reconnect_conservative,
    "diffusion": reconnect_diffusion,
    "euclidean": reconnect_euclidean,
}

#############################
## MAIN TRACKING FUNCTIONS ##
#############################


def track(
    locs,
    method="diffusion",
    search_radius=2.5,
    pixel_size_um=0.16,
    frame_interval=0.00548,
    min_I0=0.0,
    max_blinks=0,
    debug=False,
    max_spots_per_frame=None,
    reindex_unassigned=True,
    **kwargs
):
    """
    Given a dataframe with localizations, reconnect into
    trajectories.

    Each frame-frame reconnection problem is considered
    separately and sequentially. For each problem:

        1. Figure out which localizations lie within the
            the search radii of the current trajectories
        2. Identify disconnected "subproblems" in this
            trajectory-localization adjacency map
        3. Solve all of the subproblems by a method
            specified by the *method* kwarg
        4. Update the trajectories and proceed to the
            next frame

    The result is an assignment of each localization to a
    trajectory index. Localizations that were not reconnected
    into a trajectory for whatever reason are assigned a
    trajectory index of -1.

    args
    ----
        locs            :   pandas.DataFrame, set of localizations
        method          :   str, the tracking method. Currently either
                            "diffusion", "euclidean", or "conservative"
        search_radius   :   float, max jump length in um
        pixel_size_um   :   float, um per pixel
        frame_interval  :   float, seconds
        min_I0          :   float, only track spots with at least this
                            intensity
        max_blinks      :   int, number of gaps allowed
        debug           :   bool
        max_spots_per_frame
                        :   int, don't track in frames with more than
                            this number of spots
        **kwargs        :   additional keyword arguments to the tracking
                            method

    returns
    -------
        pandas.Series, trajectory indices for each localization.

    """
    # Filter on the number of spots per frame, if desired
    if not max_spots_per_frame is None:
        return track_subset(
            locs,
            [
                lambda T: filter_on_spots_per_frame(
                    T,
                    max_spots_per_frame=max_spots_per_frame,
                    filter_kernel=21,
                )
            ],
            method=method,
            search_radius=search_radius,
            pixel_size_um=pixel_size_um,
            frame_interval=frame_interval,
            min_I0=min_I0,
            max_blinks=max_blinks,
            debug=debug,
            **kwargs
        )

    # If passed an empty dataframe, do nothing
    if locs.empty:
        for c in ["trajectory", "subproblem_n_locs", "subproblem_n_traj"]:
            locs[c] = []
        return locs

    # Determine frame limits for tracking
    start_frame = int(locs["frame"].min())
    stop_frame = int(locs["frame"].max()) + 1

    # Get the reconnection method
    method_f = METHODS.get(method)

    # Sort the localizations by frame (unecessary, but easier
    # to interpret)
    locs = locs.sort_values(by="frame")

    # Assign each localization a unique index
    locs["loc_idx"] = np.arange(len(locs))

    # Convert locs to ndarray for speed
    cols = ["loc_idx", "frame", "y", "x", "I0"]
    if "I0" not in locs.columns:
        locs["I0"] = 1.0
    L = np.asarray(locs[cols])

    # Maximum tolerated traj-loc jump distance (search radius)
    search_radius_pxl = search_radius / pixel_size_um

    # Convenience function: get all of locs from one frame
    def get_locs(frame):
        return L[L[:, 1] == frame, :]

    # Convenience function: in a list of Trajectory, find
    # trajectories that have finished
    def get_finished(trajs):
        _finished = [t for t in trajs if not t.active]
        _active = [t for t in trajs if t.active]
        return _active, _finished

    # Start by grabbing the locs in the first frame and
    # initializing Trajectories from each of them
    frame_locs = get_locs(start_frame)
    active = [
        Trajectory(int(i), L, (0, 1), max_blinks) for i in frame_locs[:, 0]
    ]

    # During tracking, Trajectories are tossed between
    # three categories: "active", "new", and "completed".
    # "active" Trajectories are eligible for reconnection in
    # this frame, "new" Trajectories will become active
    # Trajectories in the next frame, and "completed" Trajectories
    # have been removed from the pool.
    new = []
    completed = []

    for fi in tqdm(
        range(start_frame + 1, stop_frame),
        total=stop_frame - (start_frame + 1),
    ):

        # # DEBUG
        # print("FRAME:\t%d" % fi)
        # print("Duplicates in active:\t", is_duplicates(active))
        # print("Duplicates in new:\t", is_duplicates(new))
        # print("Duplicates in completed:\t", is_duplicates(completed))
        # print("Completed trajectories:")
        # for i, t in enumerate(completed):
        #     print("\tTrajectory %d" % i)
        #     print(t.get_slice()[:,:2])
        # print("\n")

        # if fi > 7:
        #     break

        frame_locs = get_locs(fi)

        # If there are no locs in this frame, set all active
        # trajectories into blink
        if len(frame_locs.shape) < 2 or frame_locs.shape[0] == 0:

            # Increment blink counter
            for t in active:
                t.blink()

            # Find which trajectories are finished
            active, done = get_finished(active)
            completed += done

            # To next frame
            continue

        # If there are no active trajectories, consider starting
        # one from each localization if it passes the intensity
        # threshold
        elif len(active) == 0:

            for i in frame_locs[frame_locs[:, 4] >= min_I0, 0]:
                new.append(Trajectory(int(i), L, (0, 1), max_blinks))
            active = new
            new = []

            # To next frame
            continue

        # Otherwise, there is some combination of active trajectories
        # and localizations in this frame.
        else:

            # Calculate the adjacency graph: which localizations are
            # within the search radius of which trajectories?
            adj_g = (
                traj_loc_distance(active, frame_locs) <= search_radius_pxl
            ).astype(np.int64)

            # Break this graph into subgraphs, each of which represents
            # a separate tracking subproblem
            (
                subgraphs,
                Ti,
                Li,
                traj_singlets,
                loc_singlets,
            ) = connected_components(adj_g)

            # # DEBUG - PASSED
            # for i in range(len(Ti)):
            #     for j in [k for k in range(len(Ti)) if k != i]:
            #         for element in Ti[i]:
            #             assert element not in Ti[j]

            # # DEBUG - PASSED
            # for i in range(len(Li)):
            #     for j in [k for k in range(len(Li)) if k != i]:
            #         for element in Li[i]:
            #             assert element not in Li[j]

            # # DEBUG - a trajectory cannot be simultaneously in the
            # # singlet list and also in a subproblem group - PASSED
            # for i in traj_singlets:
            #     for j in range(len(Ti)):
            #         assert i not in Ti[j]
            # for i in loc_singlets:
            #     for j in range(len(Li)):
            #         assert i not in Li[j]

            # If a trajectory does not have localizations in its
            # search radius, set it into blink
            for ti in traj_singlets:
                active[ti].blink()
                if active[ti].active:
                    new.append(active[ti])
                else:
                    completed.append(active[ti])

            # If a localization has no nearby trajectories, start
            # a new trajectory if it passes the intensity threshold
            for li in loc_singlets:
                if frame_locs[li, 4] >= min_I0:
                    new.append(
                        Trajectory(frame_locs[li, 0], L, (0, 1), max_blinks)
                    )

            # If there are both trajectories and localizations in the
            # subproblem, reconnect according to the reconnection method
            for si, subgraph in enumerate(subgraphs):

                # Only one traj and one loc: assignment is unambiguous
                if subgraph.shape[0] == 1 and subgraph.shape[1] == 1:
                    active[Ti[si][0]].add_index(
                        frame_locs[Li[si][0], 0], (1, 1)
                    )
                    new.append(active[Ti[si][0]])

                # Otherwise, pass to the reconnection method
                else:
                    in_trajs = [active[i] for i in Ti[si]]
                    out_trajs = method_f(
                        [active[i] for i in Ti[si]],
                        frame_locs[Li[si], :],
                        L,
                        max_blinks=max_blinks,
                        pixel_size_um=pixel_size_um,
                        frame_interval=frame_interval,
                        search_radius=search_radius,
                        **kwargs
                    )

                    # Find finished trajectories
                    not_done, done = get_finished(out_trajs)
                    completed += done
                    new += not_done

            # For trajs eligible for reconnection in the next frame,
            # transfer to *active*
            active = new
            new = []

    # Finish any trajectories still running
    completed += active

    # Trajectory indices
    ids = np.full(L.shape[0], -1, dtype=np.int64)

    # Number of competing trajectories and competing localizations
    # for the subproblem in which each localization was connected
    # (1: no competition)
    subproblem_sizes_traj = np.full(L.shape[0], -1, dtype=np.int64)
    subproblem_sizes_locs = np.full(L.shape[0], -1, dtype=np.int64)

    # For each trajectory, add its information to these arrays
    for ti, t in enumerate(completed):
        indices = np.asarray(t.indices)
        T_size = [t.subproblem_shapes[i][0] for i in range(len(indices))]
        L_size = [t.subproblem_shapes[i][1] for i in range(len(indices))]
        ids[np.asarray(t.indices)] = ti
        subproblem_sizes_traj[np.asarray(t.indices)] = T_size
        subproblem_sizes_locs[np.asarray(t.indices)] = L_size

    # Assign traj index as a column in the original dataframe
    locs["trajectory"] = ids
    locs["subproblem_n_traj"] = subproblem_sizes_traj
    locs["subproblem_n_locs"] = subproblem_sizes_locs

    # For localizations unassigned to any trajectory, assign
    # unique trajectory indices
    if reindex_unassigned:
        max_index = locs["trajectory"].max() + 1
        unassigned = locs["trajectory"] == -1
        n_unassigned = unassigned.sum()
        locs.loc[unassigned, "trajectory"] = np.arange(
            max_index, max_index + n_unassigned
        )

    # If desired, return the Trajectory objects for testing
    if debug:
        return locs, completed
    else:
        return locs


def track_subset(locs, filters, **kwargs):
    """
    Run tracking on a subset of trajectories - for instance,
    only on localizations in frames that meet a particular
    criterion.

    *filters* should be a set of functions that take trajectory
    dataframes as argument and return a boolean pandas.Series
    indicating whether each localization passed or failed the
    criterion.

    Example for the *filters* argument:

        filters = [
            lambda locs: filter_on_spots_per_frame(
                locs,
                max_spots_per_frame=3,
                filter_kernel=21
            )
        ]

    This would only track localizations belonging to frames with
    3 or fewer total localizations in that frame. (After smoothing
    with a uniform kernel of width 21 frames, that is.)

    Parameters
    ----------
        locs            :   pandas.DataFrame, localizations
        filters         :   list of functions, the filters
        kwargs          :   keyword arguments to the tracking method

    Returns
    -------
        pandas.DataFrame, trajectories

    """
    # Determine which localizations pass the QC filters
    filter_pass = np.ones(len(locs), dtype=np.bool)
    for f in filters:
        filter_pass = np.logical_and(filter_pass, f(locs))

    # Track the subset of localizations that pass QC filters
    L = locs[filter_pass].copy()
    L = track(L, **kwargs)

    # Join the results with the original dataframe
    for c in [j for j in L.columns if j not in locs.columns]:
        locs[c] = L[c]
        locs[c] = (locs[c].fillna(-1)).astype(np.int64)

    return locs



#!/usr/bin/env python
"""
quot.helper.py -- low-level utilities

"""
# Paths
import os
from glob import glob 

# Numeric
import numpy as np 

# Dataframes
import pandas as pd 

# Special functions
from scipy.special import erf 

# Image processing utilities
from scipy import ndimage as ndi 

# Load *.mat format files
from scipy import io as sio 

# Caching
from functools import lru_cache 

# Warnings, to filter out warnings for a potentially dangerous
# division in rs() that is subsequent corrected
import warnings 

# Profiling
from time import time 

# from .plot import wireframe_overlay

def time_f(f):
    """
    Decorator. When executing the decorated function,
    also print the time to execute.

    args
    ----
        f   :   function

    returns
    -------
        output of f

    """
    def g(*args, **kwargs):
        t0 = time()
        r = f(*args, **kwargs)
        t1 = time()
        print("%f seconds" % (t1-t0))
        return r 
    return g

#############
## GENERAL ##
#############

def assign_methods(methods):
    """
    Decorator that designates a wrapper function for 
    other methods. Adds a new kwarg to the decorated 
    function ("method") that looks up the corresponding
    method in the *methods* dict.

    Example:
        def add(x, y):
            return x + y
        def mult(x, y):
            return x * y
        METHODS = {'add': add, 'mult': mult}

        do_operation = assign_methods(METHODS)(lambda a, b: None)

    This would be called:
        do_operation(2, 3, method='add')  # is 5
        do_operation(2, 3, method='mult') # is 6

    args
    ----
        methods     :   dict that keys strings to methods

    returns
    -------
        function, wrapper for the methods

    """ 
    def dec(f):
        def apply_method(*args, method=None, **kwargs):
            print(method)
            method_f = methods.get(method)
            if method_f is None:
                raise KeyError("Method %s not found; available methods are %s" % (
                    method, ", ".join(methods.keys())))
            return method_f(*args, **kwargs)
        return apply_method
    return dec

def stable_divide_array(N, D, zero=0.001):
    """
    Divide array N by array D, setting zeros in D
    to *zero*. Assumes nonzero D.

    args
    ----
        N, D    :   ndarrays of the same shape
        zero    :   float

    returns
    -------
        ndarray of the same shape as N and D

    """
    return N/np.clip(D, zero, np.inf)

def stable_divide_float(N, D, inf=0.0):
    """
    Divide float N by float D, returning *inf*
    if *D* is zero.

    args
    ----
        N, D    :   float
        inf     :   float

    returns
    -------
        float

    """
    return (inf if D==0.0 else N/D)

#####################
## FILE CONVERSION ##
#####################

def tracked_mat_to_csv(path, out_csv=None, pixel_size_um=0.16):
    """
    Convert a file from *Tracked.mat format, a MATLAB-based
    format for trajectories to a DataFrame format.

    args
    ----
        path        :   str, path to a *Tracked.mat file
        out_csv     :   str, file to save the result to 
        pixel_size_um:  float, the size of pixels in um

    returns
    -------
        pandas.DataFrame, the contents of the *Tracked.mat 
            file as a DataFrame

    """
    f = sio.loadmat(path)['trackedPar']
    if len(f.shape) == 2 and f.shape[0] == 1:
        f = f[0,:]

    # Total number of localizations in the file
    n_locs = sum([f[i][0].shape[0] for i in range(len(f))])

    # Output dataframe
    df = pd.DataFrame(index=np.arange(n_locs),
        columns=["y_um", "x_um", "frame", "trajectory", "time"])

    # Trajectory indices
    df["trajectory"] = np.concatenate([[i for j in range(f[i][0].shape[0])] for i in range(len(f))])

    # XY positions of each localization
    df[["y_um", "x_um"]] = np.concatenate([f[i][0] for i in range(len(f))], axis=0)

    # Frame index (correct for MATLAB off-by-1)
    df["frame"] = np.concatenate([f[i][1][:,0] for i in range(len(f))]) - 1

    # Timepoint in seconds
    df["time"] = np.concatenate([f[i][2][:,0] for i in range(len(f))])

    # Convert XY positions to pixels and correct for the MATLAB
    # off-by-1
    df[["y", "x"]] = (df[["y_um", "x_um"]] / pixel_size_um) - 1

    # Assign other columns that may be expected by other utilities
    df["error_flag"] = 0.0

    if not out_csv is None:
        df.to_csv(out_csv, index=False)

    return df

def load_tracks_dir(dirname, suffix="trajs.csv", start_frame=0,
    min_track_len=1):
    """
    Load all of the trajectory files in a target directory
    into a single pandas.DataFrame.

    args
    ----
        dirname         :   str, directory containing track CSVs
        suffix          :   str, extension for the track CSVs
        start_frame     :   int, exclude all trajectories before
                            this frame
        min_track_len   :   int, the minimum trajectory length to 
                            include

    returns
    -------
        pandas.DataFrame

    """
    # Find target files
    if os.path.isdir(dirname):
        target_csvs = glob(os.path.join(dirname, "*{}".format(suffix)))
        if len(target_csvs) == 0:
            raise RuntimeError("quot.helper.load_tracks_dir: could not find " \
                "trajectory CSVs in directory {}".format(dirname))
    elif os.path.isfile(dirname):
        target_csvs = [dirname]

    # Concatenate trajectories
    tracks = [pd.read_csv(j) for j in target_csvs]
    tracks = concat_tracks(*tracks)

    # Exclude points before the start frame
    tracks = tracks[tracks["frame"] >= start_frame]

    # Exclude trajectories that are too short
    tracks = track_length(tracks)
    if min_track_len > 1:
        tracks = tracks[tracks["track_length"] >= min_track_len]

    return tracks 


###############
## DETECTION ##
###############

def pad(I, H, W, mode='ceil'):
    """
    Pad an array with zeroes around the edges, placing 
    the original array in the center. 

    args
    ----
        I       :   2D ndarray, image to be padded
        H       :   int, height of final image
        W       :   int, width of final image
        mode    :   str, either 'ceil' or 'floor'. 'ceil'
                    yields convolution kernels that function
                    similarly to scipy.ndimage.uniform_filter.

    returns
    -------
        2D ndarray, shape (H, W)

    """
    H_in, W_in = I.shape
    out = np.zeros((H, W))
    if mode == "ceil":
        hc = np.ceil(H / 2 - H_in / 2).astype(int)
        wc = np.ceil(W / 2 - W_in / 2).astype(int)
    elif mode == "floor":
        hc = np.floor(H / 2 - H_in / 2).astype(int)
        wc = np.floor(W / 2 - W_in / 2).astype(int)
    out[hc : hc + H_in, wc : wc + W_in] = I
    return out

def label_spots(binary_img, intensity_img=None, mode="max"):
    """
    Find continuous nonzero objects in a binary image,
    returning the coordinates of the spots.

    If the objects are larger than a single pixel,
    then to find the central pixel do
        1. use the center of mass (if mode == 'centroid')
        2. use the brightest pixel (if mode == 'max')
        3. use the mean position of the binary spot
            (if img_int is not specified)

    args
    ----
        binary_img      :   2D ndarray (YX), dtype bool
        intensity_img   :   2D ndarray (YX)
        mode            :   str, 'max' or 'centroid'

    returns
    -------
        2D ndarray (n_spots, 2), dtype int64,
            the Y and X coordinate of each spot

    """
    # Find and label every nonzero object
    img_lab, n = ndi.label(binary_img, structure=np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]]))
    index = np.arange(1, n + 1)

    # Find the centers of the spots
    if intensity_img is None:
        positions = np.asarray(ndi.center_of_mass(binary_img, 
            labels=img_lab, index=index))
    elif mode == "max":
        positions = np.asarray(ndi.maximum_position(intensity_img,
            labels=img_lab, index=index))
    elif mode == "centroid":
        positions = np.asarray(ndi.center_of_mass(intensity_img,
            labels=img_lab, index=index))

    return positions.astype("int64")

def threshold_image(I, t=200.0, return_filt=False, mode='max'):
    """
    Take all spots in an image above a threshold *t*
    and return their approximate centers.

    If *return_filt* is set, the function also returns the 
    raw image and binary image. This is useful as a 
    back door when writing GUIs.

    args
    ----
        I           :   2D ndarray, image
        t           :   float, threshold
        return_filt :   bool
        mode        :   str, either 'max' or 'centroid'

    returns
    -------
        If *return_filt* is set:
        (
            2D ndarray, same as I;
            2D ndarray, binary image;
            2D ndarray, shape (n_spots, 2), the spot
                coordinates
        )
        else
            2D ndarray of shape (n_spots, 2), 
                the spot coordinates

    """
    I_bin = I > t
    pos = label_spots(I_bin, intensity_img=I, mode=mode)
    if return_filt:
        return I, I_bin, pos 
    else:
        return pos

def hollow_box_var(I, w0=15, w1=9):
    """
    Calculate the local variance of every point in a 2D image,
    returning another 2D image composed of the variances.

    Variances are calculated in a "hollow box" region around
    each point. The box is a square kernel of width *w0*; the
    hollow central region is width *w1*.

    For instance, if w0 = 5 and w1 = 3, the kernel would be

            1   1   1   1   1
            1   0   0   0   1
            1   0   0   0   1
            1   0   0   0   1
            1   1   1   1   1

    args
    ----
        I       :   2D ndarray, image
        w0      :   int, width of the kernel
        w1      :   int, width of central region

    returns
    -------
        2D ndarray

    """
    # Make the hollow box kernel
    hw0 = w0 // 2
    hw1 = w1 // 2
    kernel = np.ones((w0, w0), dtype='float64')
    kernel[hw0-hw1:hw0+hw1+1, hw0-hw1:hw0+hw1+1] = 0
    kernel = kernel / kernel.sum()

    # Calculate variance
    kernel_rft = np.fft.rfft2(pad(kernel, *I.shape))
    I_mean = np.fft.fftshift(np.fft.irfft2(np.fft.rfft2(I) * kernel_rft, s=I.shape))
    I2_mean = np.fft.fftshift(np.fft.irfft2(np.fft.rfft2(I**2) * kernel_rft, s=I.shape))
    return I2_mean - (I_mean**2)



###########################
## SUBPIXEL LOCALIZATION ##
###########################

## LOW-LEVEL SUBPIXEL LOCALIZATION FUNCTIONS

def ring_mean(I):
    """
    Take the mean of the outer ring of pixels in 
    a 2D array.

    args
    ----
        I         :   2D ndarray (YX)

    returns
    -------
        float, mean

    """
    return sum([
        I[0,:-1].mean(), 
        I[:-1,-1].mean(),
        I[-1,1:].mean(),
        I[1:,0].mean(),
    ])/4.0

def ring_var(I, ddof=0):
    """
    Take the variance of the outer ring of pixels in
    a 2D ndarray.

    args
    ----
        I       :   2D ndarray, image
        ddof    :   int, delta degrees of freedom

    returns
    -------
        float, variance estimate

    """
    return np.concatenate([
        I[0,1:],
        I[1:,-1],
        I[-1,:-1],
        I[:-1,0]
    ]).var(ddof=ddof)

def I0_is_crazy(I0, max_I0=10000):
    """
    Determine whether the intensity estimate in subpixel
    localization looks crazy. For many methods, this is 
    often the first thing to diverge.

    args
    ----
        I0  :       float, the intensity estimate for 
                    subpixel fitting
        max_I0  :   float, the maximum tolerated intensity

    returns
    -------
        bool, True if crazy 

    """
    return (I0>=0) and (I0<=max_I0)

def estimate_I0(I, y, x, bg, sigma=1.0):
    """
    Estimate the integrated intensity of a 2D integrated 
    Gaussian PSF. This model has four parameters - y, x,
    I0, and bg - so with y, x, and bg constrained, we can 
    solve for I0 directly.

    In this method, we use the brightest pixel in 
    the image to solve for I0 given the PSF model.
    This is easy but often an overestimate in the presence
    of Poisson noise.

    args
    ----
        I       :   2D ndarray, the PSF image
        y, x    :   float, center of PSF in y and x
        bg      :   float, estimated background intensity
                        per pixel (AU)
        sigma   :   float, 2D integrated Gaussian PSF 
                        radius

    returns
    -------
        float, estimate for I0

    """
    # Get the index of the brightest pixel
    ym, xm = np.unravel_index(np.argmax(I), I.shape)

    # Evaluate the intensity estimate
    return stable_divide_float(
        I[ym,xm] - bg,
        int_gauss_psf_1d(ym, y, sigma=sigma) * \
            int_gauss_psf_1d(xm, x, sigma=sigma),
        inf=np.nan 
    )

def estimate_I0_multiple_points(I, y, x, bg, sigma=1.0):
    """
    Estimate the integrated intensity of a 2D integrated 
    Gaussian PSF. This model has four parameters - y, x,
    I0, and bg - so with y, x, and bg constrained, we can 
    solve for I0 directly.

    In this method, we use the intensities at the 3x3 
    central points in the image window to obtain nine 
    estimates for I0, then average them for the final 
    estimate. This is more accurate than estimate_I0()
    but is slower. Probably overkill for a first guess.

    args
    ----
        I       :   2D ndarray, the PSF image
        y, x    :   float, center of PSF in y and x
        bg      :   float, estimated background intensity
                        per pixel (AU)
        sigma   :   float, 2D integrated Gaussian PSF 
                        radius

    returns
    -------
        float, estimate for I0

    """
    # Get the three central points in the spot window
    # and their coordinates
    wy, wx = I.shape
    hwy = wy//2
    hwx = wx//2
    Y, X = indices(wy, wx)
    sy = slice(hwy-1, hwy+2)
    sx = slice(hwx-1, hwx+2)
    obs_I = I[sy, sx].ravel()
    cen_y = Y[sy, sx].ravel()
    cen_x = X[sy, sx].ravel()

    return stable_divide_array(
        obs_I - bg,
        int_gauss_psf_1d(cen_y, y, sigma=sigma) * \
            int_gauss_psf_1d(cen_x, x, sigma=sigma)
    ).mean()

def amp_from_I(I0, sigma=1.0):
    """
    Given a 2D Gaussian PSF, return the PSF
    peak amplitude given the intensity `I0`.
    `I0` is equal to the PSF integrated above
    background, while `amp` is equal to the PSF
    evaluated at its maximum.

    args
    ----
         I0      : float, intensity estimate
         sigma   : float, width of Gaussian

    returns
    -------
        float, amplitude estimate

    """
    return I0/(2*np.pi*(sigma**2))

def estimate_snr(I, I0):
    """
    Estimate the signal-to-noise ratio of a PSF, 
    given an estimate *I0* for its intensity.

    args
    ----
        I       :   2D ndarray, the PSF image
        I0      :   float, intensity estimate

    returns
    -------
        float, SNR estimate

    """
    return stable_divide_float(
        amp_from_I(I0)**2, ring_var(I, ddof=1),
        inf=np.inf)

def invert_hessian(H, ridge=0.0001):
    """
    Invert a Hessian with ridge regularization to 
    stabilize the inversion.

    args
    ----
        H       :   2D ndarray, shape (n, n)
        ridge   :   float, regularization term

    returns
    -------
        2D ndarray, shape (n, n), the 
            inverted Hessian

    """
    D = np.diag(np.ones(H.shape[0])*ridge)
    while 1:
        try:
            H_inv = np.linalg.inv(H-D)
            return H_inv 
        except (ZeroDivisionError, np.linalg.linalg.LinAlgError):
            D *= 10 
            continue 

def check_2d_gauss_fit(img_shape, pars, max_I0=10000):
    """
    Check whether the fit parameters for a 2D symmetric 
    Gaussian with static sigma are sane.

    This includes:
        - Are the y and x coordinates inside the PSF
            window?
        - Are there negative intensities?
        - Are there crazily high intensities?

    args
    ----
        img_shape       :   (int, int), the PSF image shape
        pars            :   1D ndarray, (y, x, I0, bg), the
                            fit parameters
        max_I0          :   float, maximum tolerated intensity

    returns
    -------
        bool, True if the fit passes the checks

    """
    return all([
        pars[0]>=0,
        pars[1]>=0,
        pars[0]<img_shape[0],
        pars[1]<img_shape[1],
        pars[2]>=0,
        pars[2]<=max_I0,
        pars[3]>=0
    ])

## PSF DEFINITIONS

@lru_cache(maxsize=1)
def psf_int_proj_denom(sigma):
    """
    Convenience function to produce the denominator
    for the Gaussian functions in int_gauss_psf_1d().

    args
    ----
        sigma   :   float

    returns
    -------
        float, np.sqrt(2 * (sigma**2))

    """
    return np.sqrt(2*(sigma**2))


@lru_cache(maxsize=1)
def indices(size_y, size_x):
    """
    Convenience wrapper for np.indices.

    args
    ----
        size_y  :   int, the number of indices
                    in the y direction
        size_x  :   int, the number of indices
                    in the x direction

    returns
    -------
        (
            2D ndarray, the Y indices of each pixel;
            2D ndarray, the X indices of each pixel
        )

    """
    return np.indices((size_y, size_x))

@lru_cache(maxsize=1)
def indices_1d(size_y, size_x):
    """
    Cached convenience function to generate two
    sets of 1D indices.

    args
    ----
        size_y      :   int, the number of indices
                        in the y direction
        size_x      ;   int, the number of indices
                        in the x direction

    returns
    -------
        (
            1D ndarray, the Y indices of each pixel;
            1D ndarray, the X indices of each pixel
        )

    """
    return np.arange(size_y).astype('float64'), \
        np.arange(size_x).astype('float64')

def int_gauss_psf_1d(Y, yc, sigma=1.0):
    """
    Return a 2D integrated Gaussian PSF with unit 
    intensity, projected onto one of the axes. 

    args
    ----
        Y       :   1D ndarray, coordinates along the axis
        yc      :   float, PSF center
        sigma   :   float, Gaussian sigma

    returns
    -------
        1D ndarray, the intensities projected along
            each row of pixels

    """
    S = psf_int_proj_denom(sigma)
    return 0.5*(erf((Y+0.5-yc)/S) - \
        erf((Y-0.5-yc)/S))

def int_gauss_psf_2d(size_y, size_x, yc, xc, I0, sigma=1.0):
    """
    Return a 2D integrated Gaussian PSF with intensity I0.
    Does not include a background term.

    args
    ----
        size_y  :   int, the number of pixels in the y
                    direction
        size_x  :   int, the number of pixels in the x
                    direction
        yc      :   float, center of PSF in y
        xc      :   float, center of PSF in x
        sigma   :   float, PSF width 

    returns
    -------
        2D ndarray of shape (size_y, size_x), the 
            PSF 

    """
    Y, X = indices_1d(size_y, size_x)
    return I0*np.outer(
        int_gauss_psf_1d(Y, yc, sigma=sigma),
        int_gauss_psf_1d(X, xc, sigma=sigma),
    )

def int_gauss_psf_deriv_1d(Y, yc, sigma=1.0):
    """
    Evaluate the derivative of an integrated Gaussian
    PSF model with unit intensity projected onto 1 axis
    with respect to its axis variable.

    args
    ----
        Y       :   1D ndarray, the pixel indices at 
                    which to evaluate the PSF 
        yc      :   float, the spot center
        sigma   :   float, Gaussian sigma

    returns
    -------
        1D ndarray, the derivative of the projection
            with respect to the axis variable at 
            each pixel

    """
    A, B = psf_point_proj_denom(sigma)
    return (np.exp(-(Y-0.5-yc)**2 / A) - \
        np.exp(-(Y+0.5-yc)**2 / A)) / B

@lru_cache(maxsize=1)
def psf_point_proj_denom(sigma):
    """
    Convenience function to produce the denominator
    for the Gaussian functions in psf_point_psf_1d().

    args
    ----
        sigma   :   float, Gaussian sigma

    returns
    -------
        float, float

    """
    S = 2*(sigma**2)
    return S, np.sqrt(np.pi*S)

def point_gauss_psf_1d(Y, yc, sigma=1.0):
    """
    Evaluate a 1D point Gaussian with unit intensity
    at a set of points.

    args
    ----
        Y       :   1D ndarray, coordinates along the axis
        yc      :   float, PSF center
        sigma   :   float, Gaussian sigma

    returns
    -------
        1D ndarray, the intensities projected along
            each row of pixels

    """
    A, B = psf_point_proj_denom(sigma)
    return np.exp(-(Y-yc)**2 / A) / B 

def point_gauss_psf_2d(size_y, size_x, yc, xc, 
    I0, sigma=1.0):
    """
    Return a 2D pointwise-evaluated Gaussian PSF
    with intensity I0.

    args
    ----
        size_y      :   int, size of the PSF subwindow
                        in y 
        size_x      :   int, size of the PSF subwindow
                        in x
        yc          :   float, PSF center in y 
        xc          :   float, PSF center in x
        I0          :   float, PSF intensity 
        sigma       :   float, Gaussian sigma

    returns
    -------
        2D ndarray, the PSF model

    """
    Y, X = indices_1d(size_y, size_x)
    return I0*np.outer(
        point_gauss_psf_1d(Y, yc, sigma=sigma),
        point_gauss_psf_1d(X, xc, sigma=sigma),
    )

## CORE FITTING ROUTINES

def rs(psf_image):
    """
    Localize the center of a PSF using the radial 
    symmetry method.

    Originally conceived by the criminally underrated
    Parasarathy R Nature Methods 9, pgs 724–726 (2012).

    args
    ----
        psf_image : 2D ndarray, PSF subwindow

    returns
    -------
        float y estimate, float x estimate

    """
    # Get the size of the image frame and build
    # a set of pixel indices to match
    N, M = psf_image.shape
    N_half = N // 2
    M_half = M // 2
    ym, xm = np.mgrid[:N-1, :M-1]
    ym = ym - N_half + 0.5
    xm = xm - M_half + 0.5 
    
    # Calculate the diagonal gradients of intensities across each
    # corner of 4 pixels
    dI_du = psf_image[:N-1, 1:] - psf_image[1:, :M-1]
    dI_dv = psf_image[:N-1, :M-1] - psf_image[1:, 1:]
    
    # Smooth the image to reduce the effect of noise, at the cost
    # of a little resolution
    fdu = ndi.uniform_filter(dI_du, 3)
    fdv = ndi.uniform_filter(dI_dv, 3)
    
    dI2 = (fdu ** 2) + (fdv ** 2)
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        m = -(fdv + fdu) / (fdu - fdv)
        
    # For pixel values that blow up, instead set them to a very
    # high float
    m[np.isinf(m)] = 9e9
    
    b = ym - m * xm

    sdI2 = dI2.sum()
    ycentroid = (dI2 * ym).sum() / sdI2
    xcentroid = (dI2 * xm).sum() / sdI2
    w = dI2 / np.sqrt((xm - xcentroid)**2 + (ym - ycentroid)**2)

    # Correct nan / inf values
    w[np.isnan(m)] = 0
    b[np.isnan(m)] = 0
    m[np.isnan(m)] = 0

    # Least-squares analytical solution to the point of 
    # maximum radial symmetry, given the slopes at each
    # edge of 4 pixels
    wm2p1 = w / ((m**2) + 1)
    sw = wm2p1.sum()
    smmw = ((m**2) * wm2p1).sum()
    smw = (m * wm2p1).sum()
    smbw = (m * b * wm2p1).sum()
    sbw = (b * wm2p1).sum()
    det = (smw ** 2) - (smmw * sw)
    xc = (smbw*sw - smw*sbw)/det
    yc = (smbw*smw - smmw*sbw)/det

    # Adjust coordinates so that they're relative to the
    # edge of the image frame
    yc = (yc + (N + 1) / 2.0) - 1
    xc = (xc + (M + 1) / 2.0) - 1

    return yc, xc

def fit_ls_int_gaussian(img, guess, sigma=1.0, ridge=0.0001, 
    max_iter=20, damp=0.3, convergence=1.0e-4, divergence=1.0):
    """
    Given an observed spot, estimate the maximum likelihood
    parameters for a 2D integrated Gaussian PSF sampled with
    normally-distributed noise. Here, we use a Levenberg-
    Marquardt procedure to find the ML parameters.

    The model parameters are (y, x, I0, bg), where y, x
    is the spot center, I0 is the integrated intensity, and
    bg is the average background per pixel.

    The function also returns several parameters that are 
    useful for quality control on the fits.

    args
    ----
        img     :   2D ndarray, the observed PSF 
        guess   :   1D ndarray of shape (4), the initial
                    guess for y, x, I0, and bg
        sigma   :   float, Gaussian PSF width 
        ridge   :   float, initial regularization term
                    for inversion of the Hessian
        max_iter:   int, the maximum number of iterations
                    tolerated
        damp    :   float, damping factor for update vector.
                    Larger means faster convergence but
                    also more unstable.
        convergence :   float, the criterion for fit
                    convergence. Only y and x are tested
                    for convergence.
        divergence  :   float, the criterion for fit
                    divergence. Fitting is terminated
                    when this is reached. CURRENTLY NOT
                    IMPLEMENTED.

    returns
    -------
        (
            1D ndarray, the final parameter estimate;

            1D ndarray, the estimated error in each
                parameter (square root of the inverse
                Hessian diagonal);

            float, the Hessian determinant;

            float, the root mean squared deviation of the
                final PSF model from the observed PSF;

            int, the number of iterations
        )

    """
    n_pixels = img.shape[0] * img.shape[1]

    # 1D indices for each pixel
    index_y, index_x = indices_1d(*img.shape)

    # Update vector
    update = np.ones(4, dtype='float64')

    # Current parameter set
    pars = guess.copy()

    # Continue improving guess until convergence
    iter_idx = 0
    while iter_idx<max_iter and (np.abs(update[:2])>convergence).any():

        # Evaluate the PSF model and its derivatives projected
        # onto each of the two axes
        proj_y = int_gauss_psf_1d(index_y, pars[0], sigma=sigma)
        proj_x = int_gauss_psf_1d(index_x, pars[1], sigma=sigma)

        # Evaluate the derivatives of the model with respect to
        # the Y and X axies
        dproj_y_dy = int_gauss_psf_deriv_1d(index_y, pars[0], sigma=sigma)
        dproj_x_dx = int_gauss_psf_deriv_1d(index_x, pars[1], sigma=sigma)

        # Evaluate the model Jacobian
        J = np.asarray([
            (pars[2] * np.outer(dproj_y_dy, proj_x)).ravel(),
            (pars[2] * np.outer(proj_y, dproj_x_dx)).ravel(),
            (np.outer(proj_y, proj_x)).ravel(),
            np.ones(n_pixels)
        ]).T 

        # Evaluate the model 
        model = pars[2] * J[:,2] + pars[3]

        # Estimate the noise variance
        devs = img.ravel() - model 
        sig2 = (devs**2).mean()

        # Get the gradient of the log-likelihood
        grad = (J.T * devs).sum(axis=1) / sig2 

        # Evaluate and invert the Hessian
        H_inv = invert_hessian(-J.T@J/sig2, ridge=ridge)

        # Get the update vector
        update[:] = -H_inv @ grad

        # Apply the update to parameter estimate 
        pars = pars + damp * update 
        iter_idx += 1

    # Estimate the error with the inverse Hessian
    err = np.sqrt(np.diagonal(-H_inv))

    # Evaluate the Hessian determinant, which is 
    # also useful as a metric of error
    return pars, err, np.linalg.det(H_inv), np.sqrt(sig2), iter_idx

def fit_ls_point_gaussian(img, guess, sigma=1.0, ridge=0.0001, 
    max_iter=20, damp=0.3, convergence=1.0e-4, divergence=1.0):
    """
    Given an observed spot, estimate the maximum likelihood
    parameters for a 2D pointwise-evaluated Gaussian PSF 
    sampled with normally-distributed noise. This function
    uses a Levenberg-Marquardt procedure to find the ML 
    parameters.

    The underlying model for the pointwise-evaluated and 
    integrated Gaussian PSFs is identical - a symmetric 2D
    Gaussian. The distinction is how they handle sampling
    on discrete pixels:

        - the point Gaussian takes the value on each pixel
            to be equal to the Gaussian evaluated at the 
            center of the pixel

        - the integrated Gaussian takes the value on each 
            pixel to be equal to the Gaussian integrated
            across the whole area of that pixel

    Integrated Gaussians are more accurate and less prone
    to edge biases, but the math is slightly more 
    complicated and fitting is slower as a result.

    The model parameters are (y, x, I0, bg), where y, x
    is the spot center, I0 is the integrated intensity, and
    bg is the average background per pixel.

    The function also returns several parameters that are 
    useful for quality control on the fits.

    args
    ----
        img     :   2D ndarray, the observed PSF 
        guess   :   1D ndarray of shape (4), the initial
                    guess for y, x, I0, and bg
        sigma   :   float, Gaussian PSF width 
        ridge   :   float, initial regularization term
                    for inversion of the Hessian
        max_iter:   int, the maximum number of iterations
                    tolerated
        damp    :   float, damping factor for update vector.
                    Larger means faster convergence but
                    also more unstable.
        convergence :   float, the criterion for fit
                    convergence. Only y and x are tested
                    for convergence.
        divergence  :   float, the criterion for fit
                    divergence. Fitting is terminated
                    when this is reached. CURRENTLY NOT
                    IMPLEMENTED.

    returns
    -------
        (
            1D ndarray, the final parameter estimate;

            1D ndarray, the estimated error in each
                parameter (square root of the inverse
                Hessian diagonal);

            float, the Hessian determinant;

            float, the root mean squared deviation of the
                final PSF model from the observed PSF;

            int, the number of iterations
        )

    """
    S2 = sigma**2
    n_pixels = img.shape[0] * img.shape[1]

    # 1D indices for each pixel
    index_y, index_x = indices_1d(*img.shape)

    # Update vector
    update = np.ones(4, dtype='float64')

    # Current parameter set
    pars = guess.copy()

    # Continue improving guess until convergence
    iter_idx = 0
    while iter_idx<max_iter and (np.abs(update[:2])>convergence).any():

        # Evaluate a unit intensity Gaussian projected onto each
        # axis
        psf_1d_y = point_gauss_psf_1d(index_y, pars[0], sigma=sigma)
        psf_1d_x = point_gauss_psf_1d(index_x, pars[1], sigma=sigma)

        # Evaluate the derivatives of the model with respect to
        # the Y and X axes
        dpsf_1d_y_dy = (index_y-pars[0]) * psf_1d_y / S2
        dpsf_1d_x_dx = (index_x-pars[1]) * psf_1d_x / S2 

        # Evaluate the model Jacobian
        J = np.asarray([
            (pars[2] * np.outer(dpsf_1d_y_dy, psf_1d_x)).ravel(),
            (pars[2] * np.outer(psf_1d_y, dpsf_1d_x_dx)).ravel(),
            (np.outer(psf_1d_y, psf_1d_x)).ravel(),
            np.ones(n_pixels)
        ]).T 

        # Evaluate the model 
        model = pars[2] * J[:,2] + pars[3]

        # Estimate the noise variance
        devs = img.ravel() - model 
        sig2 = (devs**2).mean()

        # Get the gradient of the log-likelihood under a
        # normally-distributed noise model
        grad = (J.T * devs).sum(axis=1) / sig2 

        # Evaluate and invert the Hessian
        H_inv = invert_hessian(-J.T@J/sig2, ridge=ridge)

        # Get the update vector
        update[:] = -H_inv @ grad

        # Apply the update to parameter estimate 
        pars = pars + damp * update 
        iter_idx += 1

    # Estimate the error with the inverse Hessian
    err = np.sqrt(np.diagonal(-H_inv))

    # Evaluate the Hessian determinant, which is 
    # also useful as a metric of error
    return pars, err, np.linalg.det(H_inv), np.sqrt(sig2), iter_idx

def fit_poisson_int_gaussian(img, guess, sigma=1.0, ridge=0.0001, 
    max_iter=20, damp=0.3, convergence=1.0e-4, divergence=1.0):
    """
    Given an observed spot, estimate the maximum likelihood
    parameters for a 2D integrated Gaussian PSF model sampled
    with Poisson noise, using a Levenberg-Marquardt procedure.

    While LM with Poisson noise is a little slower than LS
    routines, it is the most accurate model for the noise on 
    EMCCD cameras.

    The model parameters are (y, x, I0, bg), where y, x
    is the spot center, I0 is the integrated intensity, and
    bg is the average background per pixel.

    The function also returns several parameters that are 
    useful for quality control on the fits.

    args
    ----
        img     :   2D ndarray, the observed PSF 
        guess   :   1D ndarray of shape (4), the initial
                    guess for y, x, I0, and bg
        sigma   :   float, Gaussian PSF width 
        ridge   :   float, initial regularization term
                    for inversion of the Hessian
        max_iter:   int, the maximum number of iterations
                    tolerated
        damp    :   float, damping factor for update vector.
                    Larger means faster convergence but
                    also more unstable.
        convergence :   float, the criterion for fit
                    convergence. Only y and x are tested
                    for convergence.
        divergence  :   float, the criterion for fit
                    divergence. Fitting is terminated
                    when this is reached. CURRENTLY NOT
                    IMPLEMENTED.

    returns
    -------
        (
            1D ndarray, the final parameter estimate;

            1D ndarray, the estimated error in each
                parameter (square root of the inverse
                Hessian diagonal);

            float, the Hessian determinant;

            float, the root mean squared deviation of the
                final PSF model from the observed PSF;

            int, the number of iterations
        )

    """
    n_pixels = img.shape[0] * img.shape[1]
    img_ravel = img.ravel()

    # 1D indices for each pixel
    index_y, index_x = indices_1d(*img.shape)

    # Update vector
    update = np.ones(4, dtype='float64')

    # Current parameter set
    pars = guess.copy()

    # Instantiate the Hessian
    H = np.empty((4, 4), dtype='float64')

    # Continue improving guess until convergence
    iter_idx = 0
    while iter_idx<max_iter and (np.abs(update[:2])>convergence).any():

        # Evaluate the PSF model and its derivatives projected
        # onto each of the two axes
        proj_y = int_gauss_psf_1d(index_y, pars[0], sigma=sigma)
        proj_x = int_gauss_psf_1d(index_x, pars[1], sigma=sigma)

        # Evaluate the derivatives of the model with respect to
        # the Y and X axies
        dproj_y_dy = int_gauss_psf_deriv_1d(index_y, pars[0], sigma=sigma)
        dproj_x_dx = int_gauss_psf_deriv_1d(index_x, pars[1], sigma=sigma)

        # Evaluate the model Jacobian
        J = np.asarray([
            (pars[2] * np.outer(dproj_y_dy, proj_x)).ravel(),
            (pars[2] * np.outer(proj_y, dproj_x_dx)).ravel(),
            (np.outer(proj_y, proj_x)).ravel(),
            np.ones(n_pixels)
        ]).T 

        # Evaluate the model 
        model = pars[2] * J[:,2] + pars[3]

        # if debug:
        #     print('iter: %d' % iter_idx)
        #     wireframe_overlay(img, model.reshape(img.shape))

        # Get the gradient of the log-likelihood
        grad = (J.T * ((img_ravel/model) - 1.0)).sum(axis=1)

        # Evaluate and invert the Hessian
        H_factor = -(img_ravel / (model**2))
        for i in range(4):
            for j in range(i, 4):
                H[i,j] = (H_factor * J[:,i] * J[:,j]).sum()
                if j > i:
                    H[j,i] = H[i,j]
        H_inv = invert_hessian(H, ridge=ridge)

        # Get the update vector
        update[:] = -H_inv @ grad

        # Apply the update to parameter estimate 
        pars = pars + damp * update 
        iter_idx += 1

    # if debug:
    #     print('final:')
    #     wireframe_overlay(img, model.reshape(img.shape))

    # Estimate the error with the inverse Hessian
    err = np.sqrt(np.diagonal(-H_inv))

    # Get the root mean squared deviation of the final
    # model from the observed spot
    rmse = ((model - img_ravel)**2).mean()

    # Evaluate the Hessian determinant, which is 
    # also useful as a metric of error
    return pars, err, np.linalg.det(H_inv), rmse, iter_idx

########################
## TRACKING UTILITIES ##
########################

def connected_components(semigraph):
    """
    Find independent connected subgraphs in a bipartite graph 
    (aka semigraph) by a floodfill procedure.

    The semigraph is a set of edges betwen two sets of nodes-
    one set corresponding to each y-index, and the other 
    corresponding to each x-index. Edges can only exist between
    a y-node and an x-node, so the semigraph is conveniently
    represented as a binary (0/1 values) 2D matrix.

    The goal of this function is to identify independent 
    subgraphs in the original semigraph. If I start on a y-node
    or x-node for one of these independent subgraphs and walk
    along edges, I can reach any point in that subgraph but no
    points in other subgraphs.

    Each subgraph can itself be represented by a 2D ndarray 
    along with a reference to the specific y-nodes and x-dnoes
    in the original semigraph corresponding to that subgraph.

    For the purposes of tracking, it is also convenient to 
    separate y-nodes and x-nodes that are NOT connected by any
    edges to other nodes. These are returned as separate ndarrays:
    y_without_x and x_without_y.
    
    Parameters
    ----------
        semigraph       :   2D ndarray, with only 0 and 1 values
    
    Returns
    -------
    (
        subgraphs           :   list of 2D ndarray
        subgraph_y_indices  :   list of 1D ndarray, the set of 
                                y-nodes corresponding to each 
                                subgraph
        subgraph_x_indices  :   list of 1D ndarray, the set of 
                                x-nodes corresponding to each 
                                subgraph
        y_without_x         :   1D ndarray, the y-nodes that are 
                                not connected to any x-nodes
        x_without_y         :   1D ndarray, the x-nodes that are 
                                not connected to any y-nodes
    )

    """
    # The set of all y-nodes (corresponding to y-indices in the semigraph)
    y_indices = np.arange(semigraph.shape[0]).astype("int64")

    # The set of all x-nodes (corresponding to x-indices in the semigraph)
    x_indices = np.arange(semigraph.shape[1]).astype("int64")

    # Find y-nodes that don't connect to any x-node,
    # and vice versa
    where_y_without_x = semigraph.sum(axis=1) == 0
    where_x_without_y = semigraph.sum(axis=0) == 0
    y_without_x = y_indices[where_y_without_x]
    x_without_y = x_indices[where_x_without_y]

    # Consider the remaining nodes, which have at least one edge
    # to a node of the other class
    semigraph = semigraph[~where_y_without_x, :]
    semigraph = semigraph[:, ~where_x_without_y]
    y_indices = y_indices[~where_y_without_x]
    x_indices = x_indices[~where_x_without_y]

    # For the remaining nodes, keep track of (1) the subgraphs
    # encoding connected components, (2) the set of original y-indices
    # corresponding to each subgraph, and (3) the set of original x-
    # indices corresponding to each subgraph
    subgraphs = []
    subgraph_y_indices = []
    subgraph_x_indices = []

    # Work by iteratively removing independent subgraphs from the
    # graph. The list of nodes still remaining are kept in
    # *unassigned_y* and *unassigned_x*
    unassigned_y, unassigned_x = (semigraph == 1).nonzero()

    # The current index is used to floodfill the graph with that
    # integer. It is incremented as we find more independent subgraphs.
    current_idx = 2

    # While we still have unassigned nodes
    while len(unassigned_y) > 0:

        # Start the floodfill somewhere with an unassigned y-node
        semigraph[unassigned_y[0], unassigned_x[0]] = current_idx

        # Keep going until subsequent steps of the floodfill don't
        # pick up additional nodes
        prev_nodes = 0
        curr_nodes = 1
        while curr_nodes != prev_nodes:
            # Only floodfill along existing edges in the graph
            where_y, where_x = (semigraph == current_idx).nonzero()

            # Assign connected nodes to the same subgraph index
            semigraph[where_y, :] *= current_idx
            semigraph[semigraph > current_idx] = current_idx
            semigraph[:, where_x] *= current_idx
            semigraph[semigraph > current_idx] = current_idx

            # Correct for re-finding the same nodes and multiplying
            # them more than once (implemented in the line above)
            # semigraph[semigraph > current_idx] = current_idx

            # Update the node counts in this subgraph
            prev_nodes = curr_nodes
            curr_nodes = (semigraph == current_idx).sum()
        current_idx += 1

        # Get the local indices of the y-nodes and x-nodes (in the context
        # of the remaining graph)
        where_y = np.unique(where_y)
        where_x = np.unique(where_x)

        # Use the local indices to pull this subgraph out of the
        # main graph
        subgraph = semigraph[where_y, :]
        subgraph = subgraph[:, where_x]

        # Save the subgraph
        if not (subgraph.shape[0] == 0 and subgraph.shape[0] == 0):
            subgraphs.append(subgraph)

            # Get the original y-nodes and x-nodes that were used in this
            # subgraph
            subgraph_y_indices.append(y_indices[where_y])
            subgraph_x_indices.append(x_indices[where_x])

        # Update the list of unassigned y- and x-nodes
        unassigned_y, unassigned_x = (semigraph == 1).nonzero()

    return subgraphs, subgraph_y_indices, subgraph_x_indices, y_without_x, x_without_y

def concat_tracks(*tracks):
    """
    Join some trajectory dataframes together into a larger dataframe,
    while preserving uniqe trajectory indices.

    args
    ----
        tracks      :   pandas.DataFrame with the "trajectory" column

    returns
    -------
        pandas.DataFrame, the concatenated trajectories

    """
    n = len(tracks)

    # Sort the tracks dataframes by their size. The only important thing
    # here is that if at least one of the tracks dataframes is nonempty,
    # we need to put that one first.
    df_lens = [len(t) for t in tracks]
    try:
        tracks = [t for _, t in sorted(zip(df_lens, tracks))][::-1]
    except ValueError:
        pass

    # Iteratively concatenate each dataframe to the first while 
    # incrementing the trajectory index as necessary
    out = tracks[0].assign(dataframe_index=0)
    c_idx = out["trajectory"].max() + 1

    for t in range(1, n):

        # Get the next set of trajectories and keep track of the origin
        # dataframe
        new = tracks[t].assign(dataframe_index=t)

        # Ignore negative trajectory indices (facilitating a user filter)
        new.loc[new["trajectory"]>=0, "trajectory"] += c_idx 

        # Increment the total number of trajectories
        c_idx = new["trajectory"].max() + 1

        # Concatenate
        out = pd.concat([out, new], ignore_index=True, sort=False)

    return out 
    
def track_length(tracks):
    """
    Generate a new column with the trajectory length in frames.

    args
    ----
        tracks  :   pandas.DataFrame

    returns
    -------
        pandas.DataFrame, the input dataframe with a new 
            column "track_length"

    """
    if "track_length" in tracks.columns:
        tracks = tracks.drop("track_length", axis=1)
    tracks = tracks.join(
        tracks.groupby("trajectory").size().rename("track_length"),
        on="trajectory"
    )
    return tracks 

#######################
## MASKING UTILITIES ##
#######################

def get_edges(bin_img):
    """
    Given a binary image that is False outside of an object and True
    inside of it, return another binary image that is True for points
    in the original image that border a False pixel, and False otherwise.

    args
    ----
        bin_img         :   2D ndarray, dtype bool

    returns
    -------
        2D ndarray, dtype bool, shape shape as bin_img

    """
    assert bin_img.dtype == np.bool, "get_edges: must be bool input"
    out = np.zeros(bin_img.shape, dtype="bool")

    # Find the edges of the binary mask
    out[1:, :] = out[1:, :] | (bin_img[1:, :] & ~bin_img[:-1, :])
    out[:, 1:] = out[:, 1:] | (bin_img[:, 1:] & ~bin_img[:, :-1])
    out[:-1, :] = out[:-1, :] | (bin_img[:-1, :] & ~bin_img[1:, :])
    out[:, :-1] = out[:, :-1] | (bin_img[:, :-1] & ~bin_img[:, 1:])

    # Figure out where the object in the binary image intersects
    # the edge of the image frame
    out[0, :] = out[0, :] | bin_img[0, :]
    out[:, 0] = out[:, 0] | bin_img[:, 0]
    out[-1, :] = out[-1, :] | bin_img[-1, :]
    out[:, -1] = out[:, -1] | bin_img[:, -1]

    return out

def get_ordered_mask_points(mask, max_points=100):
    """
    Given the edges of a two-dimensional binary mask, construct a line
    around the mask.

    args
    ----
        mask        :   2D ndarray, dtype bool, mask edges as 
                        returned by get_edges
        max_points  :   int, the maximum number of points tolerated
                        in the final mask. If the number of points 
                        exceeds this, the points are repeatedly 
                        downsampled until there are fewer than 
                        max_points.

    returns
    -------
        2D ndarray of shape (n_points, 2), the points belonging
            to this ROI

    """
    # Get the X and Y coordinates of all points in the mask edge
    points = np.asarray(mask.nonzero()).T

    # Keep track of which points we've included so far
    included = np.zeros(points.shape[0], dtype=np.bool)

    # Start at the first point
    ordered_points = np.zeros(points.shape, dtype=points.dtype)
    ordered_points[0,:] = points[0,:]
    included[0] = True 

    # Index of the current point
    c = 0
    midx = 0

    # Find the closest point to the current point
    while c < points.shape[0]-1:

        # Compute distances to every other point
        distances = np.sqrt(((points[midx,:]-points)**2).sum(axis=1))

        # Set included points to impossible distances
        distances[included] = np.inf 

        # Among the points not yet included in *ordered_points*,
        # choose the one closest to the current point
        midx = np.argmin(distances)

        # Add this point to the set of ordered points
        ordered_points[c+1,:] = points[midx, :]

        # Mark this point as included
        included[midx] = True 

        # Increment the current point counter
        c += 1

    # Downsample until there are fewer than max_points
    while ordered_points.shape[0] > max_points:
        ordered_points = ordered_points[::2,:]

    return ordered_points





import numpy as np
from sklearn.neighbors import radius_neighbors_graph
import pandas as pd


def density_filtering(
    pos: pd.DataFrame, radius: float = 0.2, n_neighbors: int = None
) -> pd.Series:
    # Filtering based on number of neighbors in radius
    A = radius_neighbors_graph(pos[["x", "y"]].values, radius=radius)
    N = np.reshape(np.array(A.sum(axis=0)), (-1,))
    if n_neighbors is None:
        n_neighbors = np.quantile(N[N > 0], 0.01)
    return pd.Series(index=pos.index, data=N >= n_neighbors)



from .base import *
from ...image_tools.localization import sliding_window_filter


class WindowPercentileFilter(MoviePreProcessor):

    widget_types = {"percentile": "FloatSpinBox", "window_size": "SpinBox"}
    widget_options = {
        "percentile": {
            "step": 1.0,
            "tooltip": "percentile of the pixel intensity values in the window which will be considered as the ground level.",
            "min": 0.0,
            "label": "Percentile",
        },
        "window_size": {
            "step": 50,
            "tooltip": "Size of the windows along which quantiles are computed",
            "label": "Window",
            "min": 100,
        },
    }

    def __init__(self, percentile: float = 3.0, window_size: int = 100):
        self.percentile = percentile
        self.window_size = window_size

    def preprocess(self, mov: da.Array) -> da.Array:

        return sliding_window_filter(
            data=mov, percentile=self.percentile, window_size=self.window_size
        )

    @property
    def name(self):
        return "Local percentile filtering"

    @property
    def action_name(self):
        return "Filter"
